import re
from typing import Optional, Union
import rdflib.namespace
from rdflib.namespace import Namespace, OWL
from rdflib import URIRef
import warnings
from querent.graph.errors import (
    InvalidURIException,
    PrefixNotFoundException,
    SplitURIWithUnknownPrefix,
)
from querent.graph.utils import URI

URI_ABBR_PATTERN = re.compile(r"^(?:([^:]*):)?([^:]+)$")


SCHEMA = Namespace("http://schema.org/")
DIG = Namespace("http://purl.org/ontology/bibo/")

URI_PATTERN = re.compile(r"^http:|^urn:|^info:|^ftp:|^https:")
URI_ABBR_PATTERN = re.compile(r"^(?:([^:]*):)?([^:]+)$")


class NamespaceManager(rdflib.namespace.NamespaceManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.graph.namespace_manager = self

    def parse_uri(self, text: Union[str, URI]) -> URIRef:
        """
        Parse input text into URI

        :param text: can be one of
              1. URI, directly return
              2. prefix:name, query namespace for prefix, return expanded URI
              3. name, use default namespace to expand it and return it
        :return: URIRef
        """
        if self.check_uriref(text):
            return self.check_uriref(text)
        elif isinstance(text, str):
            text = text.strip()
            m = URI_ABBR_PATTERN.match(text)
            if m:
                prefix, name = m.groups()
                base = self.store.namespace(prefix if prefix else "")
                if not base:
                    raise PrefixNotFoundException("Prefix: %s", prefix)
                return URIRef(base + name)
        elif isinstance(text, URI):
            return self.parse_uri(text.value)
        raise InvalidURIException("Invalid URI: %s", text)

    def bind(self, prefix: str, namespace: str, override=True, replace=True):
        """
        bind a given namespace to the prefix, forbids same prefix with different namespace

        :param prefix:
        :param namespace:
        :param override: if override, rebind, even if the given namespace is already bound to another prefix.
        :param replace: if replace, replace any existing prefix with the new namespace
        """
        namespace = URIRef(str(namespace))
        # When documenting explain that override only applies in what cases
        if prefix is None:
            prefix = ""
        bound_namespace = self.store.namespace(prefix)
        # Check if the bound_namespace contains a URI and if so convert it into a URIRef for
        # comparison. This is to prevent duplicate namespaces with the same URI.
        if bound_namespace:
            bound_namespace = URIRef(bound_namespace)
        if bound_namespace and bound_namespace != namespace:
            # prefix already in use for different namespace
            if replace:
                self.store.bind(prefix, namespace)
            else:
                warnings.warn(
                    "Prefix ({}, {}) already defined, if want to replace it, set flag replace to True".format(
                        prefix if prefix else None, self.store.namespace(prefix)
                    )
                )
        else:
            bound_prefix = self.store.prefix(namespace)
            if bound_prefix is None:
                self.store.bind(prefix, namespace)
            elif bound_prefix == prefix:
                pass  # already bound
            else:
                if override or bound_prefix.startswith("_"):
                    self.store.bind(prefix, namespace)

    @staticmethod
    def check_uriref(text: Union[str, URI]) -> Optional[URIRef]:
        """
        Check if the input text is likely to be an URIRef and return None or URIRef
        """
        if isinstance(text, URIRef):
            return text
        if isinstance(text, URI):
            text = text.value
        if isinstance(text, str):
            text = text.strip()
            if URI_PATTERN.match(text.strip()):
                return URIRef(text)

    def split_uri(self, uri: str):
        """
        Overwrite rdflib's implementation which has a lot of issues
        """
        ns = ""
        for prefix, namespace in self.store.namespaces():
            if uri.startswith(namespace) and len(namespace) > len(ns):
                ns = namespace
        if ns:
            return ns, uri[len(ns) :]
        raise SplitURIWithUnknownPrefix()

    def bind_for_schema_config(self):
        """
        Bind must-have namespaces for master config, note RDF and XSD are already bound
        """
        self.bind("owl", OWL)
        self.bind("", "http://isi.edu/default-ns/")

    def compute_qname(self, uri, generate=True):
        namespace, name = self.split_uri(uri)
        namespace = URIRef(namespace)
        prefix = self.store.prefix(namespace)
        if prefix is None:
            if not generate:
                raise Exception("No known prefix for %s and generate=False")
            num = 1
            while 1:
                prefix = "ns%s" % num
                if not self.store.namespace(prefix):
                    break
                num += 1
            self.bind(prefix, namespace)
        return prefix, namespace, name



import asyncio
import pytest
import os
from querent.collectors.collector_resolver import CollectorResolver
from querent.config.collector.collector_config import DriveCollectorConfig
from querent.common.uri import Uri
import uuid
from dotenv import load_dotenv

load_dotenv()


@pytest.fixture
def drive_config():
    return DriveCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "drive_refresh_token": os.getenv("DRIVE_REFRESH_TOKEN"),
            "drive_token": os.getenv("DRIVE_TOKEN"),
            "drive_scopes": os.getenv("DRIVE_SCOPES"),
            "chunk_size": 1024 * 1024,
            "drive_client_id": os.getenv("DRIVE_CLIENT_ID"),
            "drive_client_secret": os.getenv("DRIVE_CLIENT_SECRET"),
            "specific_file_type": "application/pdf",
            "folder_to_crawl": "1BtLKXcYBrS16CX0R4V1X7Y4XyO9Ct7f8",
            "name": "Drive-config",
            "config": {},
            "uri": "drive://",
        }
    )


@pytest.mark.asyncio
async def test_google_drive_collector(drive_config):
    uri = Uri("drive://")
    resolver = CollectorResolver()
    collector = resolver.resolve(uri, drive_config)
    assert collector is not None
    await collector.connect()

    async def poll_and_print():
        counter = 0
        async for result in collector.poll():
            assert result is not None
            chunk = result.unwrap()

            if chunk is not None:
                counter += 1
        assert counter == 9

    await poll_and_print()


if __name__ == "__main__":
    asyncio.run(test_google_drive_collector())



import os
from pydantic import BaseModel, Field
from typing import List, Dict, Any
import os

from querent.config.engine.engine_config import EngineConfig

class LLM_Config(BaseModel):
    name: str = "LLMEngine"
    description: str = "An engine for NER and knowledge graph operations."
    version: str = "0.0.1"
    logger: str = "LLM.engine_config"
    ner_model_name: str = "dbmdz/bert-large-cased-finetuned-conll03-english"
    rel_model_type: str = 'llama'
    rel_model_path: str = './tests/llama-2-7b-chat.Q5_K_M.gguf'
    grammar_file_path: str = './querent/kg/rel_helperfunctions/json.gbnf'
    emb_model_name: str = 'sentence-transformers/all-MiniLM-L6-v2'
    user_context: str = Field(default="""Please analyze the provided context and two entities. Use this information to answer the users query below.
Context: {context}
Entity 1: {entity1} and Entity 2: {entity2}
Query: In a semantic triple (Subject, Predicate & Object) framework, determine which of the above entity is the subject and which is the object based on the context along with the predicate between these entities. Please also identify the subject type, object type & predicate type.
Answer:""")
    enable_filtering: bool = False
    filter_params: dict = Field(default_factory=lambda: {
        'score_threshold': 0.6,
        'attention_score_threshold': 0.1,
        'similarity_threshold': 0.5,
        'min_cluster_size': 5,
        'min_samples': 3,
        'cluster_persistence_threshold': 0.4
    })
    sample_entities: List[str] = Field(default_factory=list, description="List of sample entities")
    fixed_entities: List[str] = Field(default_factory=list, description="List of fixed entities")
    is_confined_search: bool = False
    skip_inferences: bool = False
    fixed_relationships: List[str] = Field(default_factory=list, description="List of fixed relationships")
    sample_relationships: List[str] = Field(default_factory=list, description="List of sample relationships")
    
    def __init__(self, config_source=None, **kwargs):
        config_data = {}
        config_data.update(kwargs)
        if config_source:
            config_data = self.load_config(config_source)
        if "config" in config_data:
            config_data.update(config_data["config"])
        super().__init__(**config_data)


    @classmethod
    def load_config(cls, config_source) -> dict:
        if isinstance(config_source, dict):
            # If config source is a dictionary, use it as the initial config data
            cls.config_data = config_source
        else:
            raise ValueError("Invalid config. Must be a valid dictionary")

        env_vars = dict(os.environ)
        cls.config_data.update(env_vars)  # Update config data with environment variables
        return cls.config_data

    




import asyncio
from pathlib import Path
from typing import AsyncGenerator
import io
from google.oauth2.credentials import Credentials
from google.auth.transport.requests import Request
from googleapiclient.discovery import build
from googleapiclient.http import MediaIoBaseDownload

from querent.collectors.collector_base import Collector
from querent.collectors.collector_factory import CollectorFactory
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.uri import Uri
from querent.config.collector.collector_config import (
    CollectorBackend,
    DriveCollectorConfig,
)
from querent.common import common_errors
import requests

from querent.logging.logger import setup_logger


class DriveCollector(Collector):
    def __init__(self, config: DriveCollectorConfig):
        self.items_to_ignore = []
        self.chunk_size = config.chunk_size
        self.refresh_token = config.drive_refresh_token
        self.token = config.drive_token
        self.scopes = config.drive_scopes
        self.client_id = config.drive_client_id
        self.client_secret = config.drive_client_secret
        self.creds = None
        self.drive_service = None
        self.chunk_size = config.chunk_size
        self.specific_file_type = config.specific_file_type
        self.folder_to_crawl = config.folder_to_crawl
        self.logger = setup_logger(__name__, "DriveCollector")
        try:
            with open("./.gitignore", "r", encoding="utf-8") as gitignore_file:
                self.items_to_ignore = set(
                    gitignore_file.read().replace("/", "").replace("*", "").splitlines()
                )
        except Exception as e:
            self.items_to_ignore = []

    async def connect(self):
        try:
            self.creds = Credentials(
                token=self.token,
                refresh_token=self.refresh_token,
                scopes=[self.scopes],
                token_uri="https://oauth2.googleapis.com/token",
                client_id=self.client_id,
                client_secret=self.client_secret,
            )
            if self.creds and self.creds.expired and self.creds.refresh_token:
                self.creds.refresh(Request())

            self.drive_service = build("drive", "v3", credentials=self.creds)
        except Exception as e:
            self.logger.error(f"Error connecting to Google Drive: {e}")
            raise common_errors.ConnectionError(
                f"Failed to connect to Google Drive: {str(e)}"
            ) from e

    async def disconnect(self):
        try:
            if self.drive_service:
                # Close the Google Drive connection
                self.drive_service.close()
        except Exception as e:
            self.logger.error(f"Error disconnecting from Google Drive: {e}")

    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        try:
            query = ""
            if self.specific_file_type:
                query = f"mimeType='{self.specific_file_type}'"
            if self.folder_to_crawl:
                if self.specific_file_type:
                    query += " and "
                query = f"'{self.folder_to_crawl}' in parents"
            results = self.drive_service.files().list(q=query).execute()
            files = results.get("files", [])
            if not files:
                self.logger.info("No files found in Google Drive")
            for file in files:
                async for chunk in self.read_chunks(file["id"]):
                    yield CollectedBytes(data=chunk, file=file["name"])
                yield CollectedBytes(data=None, file=file["name"], eof=True)
        except Exception as e:
            raise common_errors.PollingError(
                f"Failed to poll Google Drive: {str(e)}"
            ) from e
        finally:
            await self.disconnect()

    async def read_chunks(self, file_id):
        file_metadata = (
            self.drive_service.files().get(fileId=file_id, fields="mimeType").execute()
        )
        mime_type = file_metadata.get("mimeType")

        if mime_type.startswith("application/vnd.google-apps."):
            if mime_type == "application/vnd.google-apps.document":
                export_mime_type = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
            elif mime_type == "application/vnd.google-apps.folder":
                return
            else:
                raise common_errors.PollingError(
                    f"Unsupported Google Docs file type: {mime_type}"
                )

            request = self.drive_service.files().export_media(
                fileId=file_id, mimeType=export_mime_type
            )
        else:
            # It's a binary file, we can proceed with normal download
            request = self.drive_service.files().get_media(fileId=file_id)

        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, request, chunksize=self.chunk_size)
        done = False
        while not done:
            status, done = downloader.next_chunk()
            if status:
                yield fh.getvalue()
                fh.seek(0)
                fh.truncate(0)


class DriveCollectorFactory(CollectorFactory):
    def backend(self) -> CollectorBackend:
        return CollectorBackend.Drive

    def resolve(self, uri: Uri, config: DriveCollectorConfig) -> Collector:
        return DriveCollector(config)



from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from typing import AsyncGenerator
import os

from querent.collectors.collector_factory import CollectorFactory
from querent.config.collector.collector_config import CollectorBackend, SlackCollectorConfig
from querent.collectors.collector_base import Collector
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.uri import Uri
from querent.common import common_errors
from querent.logging.logger import setup_logger


class SlackCollector(Collector):
    def __init__(self, config: SlackCollectorConfig):
        self.cursor = config.cursor
        self.include_all_metadata = (
            config.include_all_metadata if config.include_all_metadata else 0
        )
        self.inclusive = config.inclusive if config.inclusive else 0
        self.latest = config.latest if config.latest else 0
        self.limit = config.limit if config.limit else 100
        self.channel = config.channel_name
        self.access_token = config.access_token

        self.client = WebClient()
        self.logger = setup_logger(__name__, "SlackCollector")

    async def connect(self):
        pass

    async def disconnect(self):
        # Add your cleanup logic here if needed
        pass

    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        try:
            self.client = WebClient(token=self.access_token)
            response = self.client.conversations_join(channel=self.channel)
            if not response["ok"]:
                self.logger.error(f"Error connecting to Slack: {response['error']}")
        except SlackApiError as exc:
            raise common_errors.SlackApiError(
                f"Slack API Error: {exc.response['error']}"
            ) from exc
        while True:
            try:
                # Get list of all the conversation history
                response = self.client.conversations_history(
                    channel=self.channel, cursor=self.cursor, limit=self.limit
                )
                if response["ok"]:
                    messages = response["messages"]
                    for message in messages:
                        yield CollectedBytes(
                            file=f"slack://{self.channel}.slack",
                            data=bytes(message["text"] + "\n\n", "utf-8"),
                        )

                    if not response["has_more"]:
                        yield CollectedBytes(
                            file=f"slack://{self.channel}.slack",
                            data=None,
                            error=None,
                            eof=True,
                        )
                        break
                else:
                    raise common_errors.SlackApiError(
                        f"Slack API Error: {response['error']}"
                    )

                self.cursor = response["response_metadata"]["next_cursor"]

            except SlackApiError as exc:
                raise common_errors.SlackApiError(
                    f"Slack API Error: {exc.response['error']}"
                ) from exc


class SlackCollectorFactory(CollectorFactory):
    def __init__(self):
        pass

    def backend(self) -> CollectorBackend:
        return CollectorBackend.Slack

    def resolve(self, uri: Uri, config: SlackCollectorConfig) -> Collector:
        return SlackCollector(config)



from typing import Dict
from querent.config.graph_config import GraphConfig
from querent.graph.graph import QuerentGraph
from querent.graph.schema import KGSchema
from querent.graph.subject import Subject
from querent.graph.utils import URI, BNode, Literal


class QuerentKG(QuerentGraph):
    """
    Querent Knowledge Graph backed by a default schema and stores knowledge in RDF format.
    """

    def __init__(self, config: GraphConfig):
        super().__init__(config)
        self.schema = KGSchema(config)
        self._fork_namespace_manager()

    def _find_types(self, triples):
        """
        find type in root level
        :param triples:
        :return:
        """
        types = []
        for t in triples:
            s, p, o = t
            if self._is_rdf_type(p):
                if isinstance(o, Subject):
                    continue
                types.append(o)
        return types

    def add_subject_knowledge(self, subjects):
        """
        Add subject knowledge to the graph
        :param subjects: list of subjects
        :return:
        """

        try:
            # TODO do some schema validation for the subject
            self.add_subjects(subjects)
        except Exception as e:
            self.logger.error(f"Error adding quad {subjects} to graph: {e}")
            raise e

    @property
    def value(self) -> Dict:
        """
        Get knowledge graph object
        """
        g = {}
        for p, o in self.graph.predicate_objects():
            _, property_ = self._ns.split_uri(p)
            if property_ not in g:
                g[property_] = list()
            g[property_].append(
                {"key": self.create_key_from_value(o, property_), "value": o.toPython()}
            )
        return g

    def create_key_from_value(self, value, field_name: str):
        key = self.schema.field_type(field_name, value)
        if isinstance(key, URI):
            return key
        if isinstance(key, str) or isinstance(key, Literal):
            key = str(key).strip().lower()
        return key

    def serialize(self):
        return super().serialize()

    def _fork_namespace_manager(self):
        for prefix, ns in self.schema.ontology._ns.namespaces():
            self.bind(prefix, ns)

    def add_types(self, type_):
        s = Subject(URI(self.origin_doc.doc_id))
        p = URI("rdf:type")

        if not isinstance(type_, list):
            type_ = [type_]
        for a_type in type_:
            s.add_property(p, URI(a_type))
        self.add_subject(s)

    def validate(self):
        conforms, result_graph = self.schema.validate(self)
        return conforms, result_graph



import asyncio
from asyncio import Queue
from typing import AsyncGenerator, List, Optional
from cachetools import LRUCache, cachedmethod

from querent.collectors.collector_base import Collector
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.types.ingested_tokens import IngestedTokens
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.ingestors.base_ingestor import BaseIngestor
from querent.ingestors.email.email_ingestor import EmailIngestorFactory
from querent.ingestors.ingestor_factory import IngestorFactory, UnsupportedIngestor
from querent.ingestors.pdfs.pdf_ingestor_v1 import PdfIngestorFactory
from querent.ingestors.texts.text_ingestor import TextIngestorFactory
from querent.ingestors.audio.audio_ingestors import AudioIngestorFactory
from querent.ingestors.json.json_ingestor import JsonIngestorFactory
from querent.ingestors.images.image_ingestor import ImageIngestorFactory
from querent.ingestors.doc.doc_ingestor import DocIngestorFactory
from querent.ingestors.csv.csv_ingestor import CsvIngestorFactory
from querent.ingestors.video.video_ingestor import VideoIngestorFactory
from querent.ingestors.xlsx.xlsx_ingestor import XlsxIngestorFactory
from querent.ingestors.ppt.ppt_ingestor import PptIngestorFactory
from querent.ingestors.xml.xml_ingestor import XmlIngestorFactory
from querent.ingestors.html.html_ingestor import HtmlIngestorFactory
from querent.ingestors.github.github_ingestor import GithubIngestorFactory
from querent.ingestors.code.code_ingestor import CodeIngestorFactory
import asyncio

from querent.logging.logger import setup_logger
from querent.processors.async_processor import AsyncProcessor


class IngestorFactoryManager:
    """Factory manager"""

    PROGRAMMING_LANGUAGES = [
        "py",
        "pyw",
        "pyp",
        "js",
        "mjs",
        "java",
        "cpp",
        "h",
        "hpp",
        "c",
        "h",
        "cs",
        "rb",
        "swift",
        "php",
        "php3",
        "php4",
        "php5",
        "css",
        "go",
        "rs",
        "kt",
        "ts",
        "pl",
        "sql",
        "r",
        "m",
        "sh",
        "bash",
        "zsh",
        "dart",
        "scala",
        "groovy",
        "lua",
        "m",
        "vb",
    ]

    def __init__(
        self,
        collectors: Optional[List[Collector]] = None,
        processors: Optional[List[AsyncProcessor]] = None,
        cache_size: Optional[int] = 100,
        result_queue: Optional[Queue] = None,
    ):
        self.collectors = collectors
        self.processors = processors
        if self.processors is None:
            self.processors = []
        if self.collectors is None:
            self.collectors = []
        self.ingestor_factories = {
            IngestorBackend.PDF.value: PdfIngestorFactory(),
            IngestorBackend.TEXT.value: TextIngestorFactory(),
            IngestorBackend.MP3.value: AudioIngestorFactory(),
            IngestorBackend.WAV.value: AudioIngestorFactory(),
            IngestorBackend.JSON.value: JsonIngestorFactory(),
            IngestorBackend.JPG.value: ImageIngestorFactory(),
            IngestorBackend.PNG.value: ImageIngestorFactory(),
            IngestorBackend.DOCX.value: DocIngestorFactory(),
            IngestorBackend.DOC.value: DocIngestorFactory(),
            IngestorBackend.CSV.value: CsvIngestorFactory(),
            IngestorBackend.XLSX.value: XlsxIngestorFactory(),
            IngestorBackend.PPT.value: PptIngestorFactory(),
            IngestorBackend.PPTX.value: PptIngestorFactory(),
            IngestorBackend.XML.value: XmlIngestorFactory(),
            IngestorBackend.HTML.value: HtmlIngestorFactory(),
            IngestorBackend.MP4.value: VideoIngestorFactory(),
            IngestorBackend.GITHUB.value: GithubIngestorFactory(),
            IngestorBackend.Slack.value: TextIngestorFactory(is_token_stream=True),
            IngestorBackend.Email.value: EmailIngestorFactory(),
            IngestorBackend.Jira.value: JsonIngestorFactory(),
            # Add more mappings as needed
        }
        self.file_caches = LRUCache(maxsize=cache_size)
        self.result_queue = result_queue
        self.logger = setup_logger(__name__, "IngestorFactoryManager")

    async def get_factory(self, file_extension: str) -> IngestorFactory:
        """get_factory to match factory based on file extension"""
        if file_extension is None or file_extension == "":
            return TextIngestorFactory()
        if file_extension in self.PROGRAMMING_LANGUAGES:
            return CodeIngestorFactory()
        return self.ingestor_factories.get(
            file_extension.lower(), UnsupportedIngestor("Unsupported file extension")
        )

    async def get_ingestor(self, file_extension: str) -> Optional[BaseIngestor]:
        """get_ingestor to get factory for that extension"""
        factory = await self.get_factory(file_extension)
        return await factory.create(file_extension, self.processors)

    async def supports(self, file_extension: str) -> bool:
        """check if extension supports factory"""
        factory = await self.get_factory(file_extension)
        return factory.supports(file_extension)

    @cachedmethod(cache=lambda self: self.file_caches)
    async def ingest_file_async(
        self, file_id: str, result_queue: Optional[Queue] = None
    ):
        try:
            collected_bytes_list = self.file_caches.pop(file_id)
            file_extension = collected_bytes_list[0].extension
            ingestor = await self.get_ingestor(file_extension)
            if ingestor is not None:

                async def chunk_generator() -> AsyncGenerator[CollectedBytes, None]:
                    for chunk in collected_bytes_list:
                        if chunk.eof or chunk.error is not None:
                            break
                        yield chunk

                async for chunk_tokens in ingestor.ingest(chunk_generator()):
                    result_queue.put_nowait(chunk_tokens)
            else:
                self.logger.warning(
                    f"Unsupported file extension {file_extension} for file {collected_bytes_list[0].file}"
                )
        except Exception as e:
            self.logger.error(
                f"Error ingesting file {collected_bytes_list[0].file}: {str(e)}"
            )

    async def ingest_collector_async(
        self, collector: Collector, result_queue: Optional[Queue] = None
    ):
        """Asynchronously ingest data from a single collector."""
        async for collected_bytes in collector.poll():
            current_file = collected_bytes.file

            if current_file not in self.file_caches:
                if not collected_bytes.eof:
                    self.file_caches[current_file] = [collected_bytes]
            else:
                if not collected_bytes.eof:
                    self.file_caches[current_file].append(collected_bytes)

            # Check if this is the end of the file
            if collected_bytes.eof:
                # Try to ingest the ongoing file even if the cache is full
                try:
                    await self.ingest_file_async(current_file, result_queue)
                except Exception as e:
                    self.logger.error(f"Error ingesting file {current_file}: {str(e)}")

                # Wait for ongoing file processing to complete before checking for the next file
                while current_file in self.file_caches:
                    await asyncio.sleep(0.1)

    async def ingest_all_async(self):
        """Asynchronously ingest data from all collectors concurrently."""
        ingestion_tasks = [
            self.ingest_collector_async(collector, self.result_queue)
            for collector in self.collectors
        ]
        await asyncio.gather(*ingestion_tasks)
        await self.result_queue.put(None)



import json
import spacy
from transformers import AutoTokenizer
from querent.kg.ner_helperfunctions.fixed_predicate import FixedPredicateExtractor
from querent.common.types.ingested_images import IngestedImages
from querent.config.core.opensource_llm_config import Opensource_LLM_Config
from querent.core.transformers.relationship_extraction_llm import RelationExtractor
from querent.kg.rel_helperfunctions.contextual_predicate import process_data
from querent.kg.ner_helperfunctions.contextual_embeddings import EntityEmbeddingExtractor
from querent.kg.ner_helperfunctions.fixed_entities import FixedEntityExtractor
from querent.kg.ner_helperfunctions.ner_llm_transformer import NER_LLM
from querent.common.types.querent_event import EventState, EventType
from querent.core.base_engine import BaseEngine
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.ingested_messages import IngestedMessages
from querent.common.types.ingested_code import IngestedCode
from querent.common.types.querent_queue import QuerentQueue
from typing import Any, List, Tuple
from querent.common.types.file_buffer import FileBuffer
from querent.logging.logger import setup_logger
from querent.kg.querent_kg import QuerentKG
from querent.graph.graph import QuerentGraph
from querent.config.graph_config import GraphConfig
from querent.kg.ner_helperfunctions.attn_scores import EntityAttentionExtractor
from querent.kg.ner_helperfunctions.filter_triples import TripleFilter
from querent.config.core.llm_config import LLM_Config
from querent.kg.rel_helperfunctions.triple_to_json import TripleToJsonConverter
from querent.kg.rel_helperfunctions.embedding_store import EmbeddingStore
import time
import psutil
"""
    BERTLLM is a class derived from BaseEngine designed for processing language models, particularly focusing on named entity recognition and relationship extraction in text. It integrates various components for handling different types of input data (messages, images, code, tokens), extracting entities, filtering relevant information, and constructing knowledge graphs.

    Key functionalities include:
    - Initializing with a specific configuration for named entity recognition (NER) and language model processing.
    - Validating the presence of NER models and tokenizers.
    - Processing various types of input data like messages, images, code, and tokens.
    - Implementing methods for counting entity pairs, setting filter parameters, and processing tokens.
    - Extracting and clustering entities and relationships from the text, and converting them into graph and vector formats.
    - Handling errors and maintaining robustness in data processing.

    The class also incorporates mechanisms for filtering and clustering entities and relationships, as well as extracting embeddings and generating output in different formats.
    """


class BERTLLM(BaseEngine):
    def __init__(
        self,
        input_queue:QuerentQueue,
        config: LLM_Config
    ):  
        self.logger = setup_logger(__name__, "BERTLLM")
        super().__init__(input_queue)
        self.skip_inferences=config.skip_inferences
        if not self.skip_inferences:
            mock_config = Opensource_LLM_Config(qa_template=config.user_context,
                                                model_type = config.rel_model_type,
                                                model_path = config.rel_model_path,
                                                grammar_file_path = config.grammar_file_path,
                                                emb_model_name = config.emb_model_name)
            self.semantic_extractor = RelationExtractor(mock_config)
        self.graph_config = GraphConfig(identifier=config.name)
        self.contextual_graph = QuerentKG(self.graph_config)
        self.semantic_graph = QuerentKG(self.graph_config)
        self.file_buffer = FileBuffer()
        self.ner_tokenizer = AutoTokenizer.from_pretrained(config.ner_model_name)
        self.ner_model = NER_LLM.load_model(config.ner_model_name, "NER")
        self.ner_llm_instance = NER_LLM(provided_tokenizer=self.ner_tokenizer, provided_model=self.ner_model)
        self.nlp_model = NER_LLM.get_class_variable()
        self.create_emb = EmbeddingStore()
        self.attn_scores_instance = EntityAttentionExtractor(model=self.ner_model, tokenizer=self.ner_tokenizer)
        self.enable_filtering = config.enable_filtering
        self.filter_params = config.filter_params or {}
        self.triple_filter = None
        if self.enable_filtering:
            self.triple_filter = TripleFilter(**self.filter_params)
        self.sample_entities = config.sample_entities
        self.fixed_entities = config.fixed_entities
        if self.fixed_entities and not self.sample_entities:
            raise ValueError("If specific entities are provided, their types should also be provided.")
        if self.fixed_entities and self.sample_entities:
            self.entity_context_extractor = FixedEntityExtractor(fixed_entities=self.fixed_entities, entity_types=self.sample_entities,model = self.nlp_model)
        elif self.sample_entities:
            self.entity_context_extractor = FixedEntityExtractor(entity_types=self.sample_entities, model = self.nlp_model)
        else:
            self.entity_context_extractor = None
        self.fixed_relationships = config.fixed_relationships
        self.sample_relationships = config.sample_relationships
        if self.fixed_relationships and not self.sample_relationships:
            raise ValueError("If specific predicates are provided, their types should also be provided.")
        if self.fixed_relationships and self.sample_relationships:
            self.predicate_context_extractor = FixedPredicateExtractor(fixed_predicates=self.fixed_relationships, predicate_types=self.sample_relationships,model = self.nlp_model)
        elif self.sample_relationships:
            self.predicate_context_extractor = FixedPredicateExtractor(predicate_types=self.sample_relationships,model = self.nlp_model)
        else:
            self.predicate_context_extractor = None
        self.user_context = config.user_context
        self.isConfinedSearch = config.is_confined_search
        
 

    def validate(self) -> bool:
        return self.ner_model is not None and self.ner_tokenizer is not None

    def process_messages(self, data: IngestedMessages):
        return super().process_messages(data)
    
    def process_images(self, data: IngestedImages):
        return super().process_images(data)

    async def process_code(self, data: IngestedCode):
        return super().process_code(data)

    @staticmethod
    def validate_ingested_tokens(data: IngestedTokens) -> bool:
        if data.is_error():
            
            return False

        return True

    def count_entity_pairs(self, doc_entity_pairs):
        total_pairs = 0
        for inner_list in doc_entity_pairs:
            total_pairs += len(inner_list)
            
        return total_pairs
    
    def set_filter_params(self, **kwargs):
        self.filter_params = kwargs
        if self.triple_filter:
            self.triple_filter.set_params(**kwargs)
        else:
            self.triple_filter = TripleFilter(**kwargs)
    
    
    async def process_tokens(self, data: IngestedTokens):
        doc_entity_pairs = []
        number_sentences = 0
        try:
            if not BERTLLM.validate_ingested_tokens(data):
                    self.set_termination_event()                                      
                    return
            if data.data:
                single_string = ' '.join(data.data)
                clean_text = single_string.replace('\n', ' ')
            else:
                clean_text = data.data
             
            file, content = self.file_buffer.add_chunk(
                data.get_file_path(), clean_text
            )
            if content:
                if self.fixed_entities:
                    content = self.entity_context_extractor.find_entity_sentences(content)
                if self.fixed_relationships:
                    content = self.predicate_context_extractor.find_predicate_sentences(content)
                tokens = self.ner_llm_instance._tokenize_and_chunk(content)
                for tokenized_sentence, original_sentence, sentence_idx in tokens:
                    (entities, entity_pairs,) = self.ner_llm_instance.extract_entities_from_sentence(original_sentence, sentence_idx, [s[1] for s in tokens],self.isConfinedSearch, self.fixed_entities, self.sample_entities)
                    if entity_pairs:
                        doc_entity_pairs.append(self.ner_llm_instance.transform_entity_pairs(entity_pairs))
                    number_sentences = number_sentences + 1
            else:
                return

            if self.sample_entities:
                doc_entity_pairs = self.entity_context_extractor.process_entity_types(doc_entities=doc_entity_pairs)
            if doc_entity_pairs:
                doc_entity_pairs = self.ner_llm_instance.remove_duplicates(doc_entity_pairs)
                pairs_withattn = self.attn_scores_instance.extract_and_append_attention_weights(doc_entity_pairs)
                if self.enable_filtering == True and not self.entity_context_extractor and self.count_entity_pairs(pairs_withattn)>1 and not self.predicate_context_extractor:
                    self.entity_embedding_extractor = EntityEmbeddingExtractor(self.ner_model, self.ner_tokenizer)
                    pairs_withemb = self.entity_embedding_extractor.extract_and_append_entity_embeddings(pairs_withattn)
                else:
                    pairs_withemb = pairs_withattn
                pairs_with_predicates = process_data(pairs_withemb, file)
                if self.enable_filtering == True and not self.entity_context_extractor and self.count_entity_pairs(pairs_withattn)>1 and not self.predicate_context_extractor:
                    cluster_output = self.triple_filter.cluster_triples(pairs_with_predicates)
                    clustered_triples = cluster_output['filtered_triples']
                    cluster_labels = cluster_output['cluster_labels']
                    cluster_persistence = cluster_output['cluster_persistence']
                          
                          
                    final_clustered_triples = self.triple_filter.filter_by_cluster_persistence(pairs_with_predicates, cluster_persistence, cluster_labels)
                    if final_clustered_triples:
                        filtered_triples, reduction_count = self.triple_filter.filter_triples(final_clustered_triples)
                    else:
                        filtered_triples, _ = self.triple_filter.filter_triples(clustered_triples)
                        self.logger.error(f"Filtering in {self.__class__.__name__} producing 0 entity pairs. Filtering Disabled. ")
                else:
                    filtered_triples = pairs_with_predicates
                if not self.skip_inferences:
                    relationships = self.semantic_extractor.process_tokens(filtered_triples[:2])
                    if len(relationships) > 0:
                        embedding_triples = self.create_emb.generate_embeddings(relationships)
                        if self.sample_relationships:
                            embedding_triples = self.predicate_context_extractor.process_predicate_types(embedding_triples)
                        for triple in embedding_triples:
                            graph_json = json.dumps(TripleToJsonConverter.convert_graphjson(triple))
                            if graph_json:
                                current_state = EventState(EventType.Graph,1.0, graph_json, file)
                                await self.set_state(new_state=current_state)
                            vector_json = json.dumps(TripleToJsonConverter.convert_vectorjson(triple))
                            if vector_json:
                                current_state = EventState(EventType.Vector,1.0, vector_json, file)
                                await self.set_state(new_state=current_state)
                    else:
                        return
                else:
                    return filtered_triples, file
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to process tokens. {e}")
            raise Exception(f"An unexpected error occurred while processing tokens: {e}")



# import asyncio
# from asyncio import Queue
# import json
# from pathlib import Path
# from querent.callback.event_callback_interface import EventCallbackInterface
# from querent.collectors.fs.fs_collector import FSCollectorFactory
# from querent.common.types.ingested_tokens import IngestedTokens
# from querent.common.types.querent_event import EventState, EventType
# from querent.config.collector.collector_config import FSCollectorConfig
# from querent.common.uri import Uri
# from querent.config.core.bert_llm_config import LLM_Config
# from querent.ingestors.ingestor_manager import IngestorFactoryManager
# import pytest
# import uuid
# from querent.common.types.file_buffer import FileBuffer
# from querent.core.transformers.bert_llm import BERTLLM
# from querent.querent.resource_manager import ResourceManager
# from querent.querent.querent import Querent


# @pytest.mark.asyncio
# async def test_ingest_all_async():
#     # Set up the collectors
#     directories = [ "./tests/data/llm/pdf/"]
#     collectors = [
#         FSCollectorFactory().resolve(
#             Uri("file://" + str(Path(directory).resolve())),
#             FSCollectorConfig(config_source={
#             "id": str(uuid.uuid4()),
#             "root_path": directory,
#             "name": "Local-config",
#             "config": {},
#             "uri": "file://",
#         }),
#         )
#         for directory in directories
#     ]

#     # Set up the result queue
#     result_queue = asyncio.Queue()
#     file_buffer = FileBuffer()

#     # Create the IngestorFactoryManager
#     ingestor_factory_manager = IngestorFactoryManager(
#         collectors=collectors, result_queue=result_queue
#     )

#     # Start the ingest_all_async in a separate task
#     ingest_task = asyncio.create_task(ingestor_factory_manager.ingest_all_async())
#     resource_manager = ResourceManager()
#     bert_llm_config = LLM_Config(
#     ner_model_name="botryan96/GeoBERT",
#     enable_filtering=True,
#     filter_params={
#             'score_threshold': 0.5,
#             'attention_score_threshold': 0.1,
#             'similarity_threshold': 0.5,
#             'min_cluster_size': 5,
#             'min_samples': 3,
#             'cluster_persistence_threshold':0.1
#         }
#             ,fixed_entities = ['carbon isotope']
#             , sample_entities=['B-GeoPetro']
#             , fixed_relationships=["occurs"]
#             ,sample_relationships=["occurs", "association"]
#     )
#     llm_instance = BERTLLM(result_queue, bert_llm_config)
#     class StateChangeCallback(EventCallbackInterface):
#         async def handle_event(self, event_type: EventType, event_state: EventState):
#             assert event_state.event_type == EventType.Graph
#             triple = json.loads(event_state.payload)
#             print("triple: {}".format(triple))
#             assert isinstance(triple['subject'], str) and triple['subject']
#     llm_instance.subscribe(EventType.Graph, StateChangeCallback())
#     querent = Querent(
#         [llm_instance],
#         resource_manager=resource_manager,
#     )

#     querent_task = asyncio.create_task(querent.start())
#     await asyncio.gather(ingest_task, querent_task)

# if __name__ == "__main__":
#     asyncio.run(test_ingest_all_async())



import json
from pydantic import BaseModel
from typing import List, Tuple, Dict

"""
    A model representing contextual information about a predicate in a knowledge graph.

    Attributes:
    -----------
    context : str
        The context in which the predicate is used.
    entity1_score, entity2_score : float
        Scores associated with the two entities involved in the predicate.
    entity1_label, entity2_label : str
        Labels associated with the two entities.
    entity1_nn_chunk, entity2_nn_chunk : str
        Nearest neighbor chunks for the two entities.
    file_path : str
        Path to the file where the predicate information is stored.
    entity1_attnscore, entity2_attnscore : float
        Attention scores for the two entities.
    entity1_embedding, entity2_embedding : List[float]
        Embeddings for the two entities.

    Methods:
    --------
    from_tuple(data: Tuple[str, str, str, Dict[str, str], str]) -> 'ContextualPredicate':
        Class method to create an instance of ContextualPredicate from a tuple.
    """

class ContextualPredicate(BaseModel):
    context: str
    entity1_score: float
    entity2_score: float
    entity1_label: str
    entity2_label: str
    entity1_nn_chunk: str
    entity2_nn_chunk: str
    file_path: str
    entity1_attnscore: float
    entity2_attnscore: float
    pair_attnscore: float
    entity1_embedding: List[float]
    entity2_embedding: List[float]
    
    
    @classmethod
    def from_tuple(cls, data: Tuple[str, str, str, Dict[str, str], str]) -> 'ContextualPredicate':
        try:
            entity1_embedding = data[3].get('entity1_embedding', []) if 'entity1_embedding' in data[3] else []
            entity2_embedding = data[3].get('entity2_embedding', []) if 'entity2_embedding' in data[3] else []

            return cls(
                context=data[1],
                entity1_score=float(data[3].get('entity1_score', 0.0)),
                entity2_score=float(data[3].get('entity2_score', 0.0)),
                entity1_label=data[3].get('entity1_label'),
                entity2_label=data[3].get('entity2_label'),
                entity1_nn_chunk=data[3].get('entity1_nn_chunk'),
                entity2_nn_chunk=data[3].get('entity2_nn_chunk'),
                entity1_attnscore=data[3].get('entity1_attnscore',''),
                entity2_attnscore=data[3].get('entity2_attnscore',''), 
                pair_attnscore=data[3].get('pair_attnscore'), 
                entity1_embedding=entity1_embedding,
                entity2_embedding=entity2_embedding,
                file_path=data[4]
            )
        except Exception as e:
            raise ValueError(f"Error creating ContextualPredicate from tuple: {e}")



def process_data(data: List[List[Tuple[str, str, str, Dict[str, float], str]]], file_path: str) -> List[Tuple[str, str, str]]:
    result = []
    try:
        for inner_list in data:
            for tup in inner_list:
                if tup:  # Check if tuple is not empty
                    extended_tup = (*tup, file_path)
                    # Convert extended tuple to ContextualPredicate object and then to JSON string
                    context_metadata_str = ContextualPredicate.from_tuple(extended_tup)
                    context_metadata_str = json.dumps(context_metadata_str.dict(), ensure_ascii=False)
                    result.append((tup[0], context_metadata_str, tup[2]))
                    
        return result
    
    except Exception as e:
        raise ValueError(f"Error processing data: {e}")




import imaplib

from querent.config.collector.collector_config import EmailCollectorConfig


class ImapEmail:
    def imap_open(self, config: EmailCollectorConfig) -> imaplib.IMAP4_SSL:
        """
        Function to open an IMAP connection to the email server.

        Args:
            config (EmailCollectorConfig): The email collector config.
        Returns:
            imaplib.IMAP4_SSL: The IMAP connection.
        """
        conn = imaplib.IMAP4_SSL(
            config.imap_server,
            config.imap_port,
            config.imap_keyfile,
            config.imap_certfile,
        )
        conn.login(config.imap_username, config.imap_password)
        conn.select(config.imap_folder)
        return conn

    def adjust_imap_folder(self, imap_folder, email_sender) -> str:
        """
        Function to adjust the IMAP folder based on the email address of the sender.

        Args:
            imap_folder (str): The folder to open.
            email_sender (str): The email address of the sender.

        Returns:
            str: The adjusted IMAP folder.
        """
        if "@gmail" in email_sender.lower():
            if "sent" in imap_folder.lower():
                return '"[Gmail]/Sent Mail"'
            if "draft" in imap_folder.lower():
                return '"[Gmail]/Drafts"'
        return imap_folder



from langchain.embeddings import HuggingFaceEmbeddings
from langchain.vectorstores import FAISS
from sentence_transformers import CrossEncoder


class RAGRetriever:
    def __init__(self, faiss_index_path, emb_model_name, embedding_store, logger):
        self.faiss_index_path = faiss_index_path
        self.emb_model_name = emb_model_name
        self.cross_encoder = self.load_cross_encoder('cross-encoder/ms-marco-TinyBERT-L-2-v2')
        self.create_emb = embedding_store
        self.logger = logger
    
    def load_cross_encoder(self, model_name):
        return CrossEncoder(model_name)
    
    def build_faiss_index(self, data):
        try:
            contexts = set()
            for item in data:
                if isinstance(item, tuple) and len(item) == 3:
                    _, predicate_dict, _ = item
                    if isinstance(predicate_dict, dict) and 'context' in predicate_dict:
                        context = predicate_dict['context']
                        context = context.encode().decode('unicode_escape')
                        contexts.add(context)
            self.create_emb.create_index(texts=contexts, verbose=True)
            self.create_emb.save_index('my_FAISS_index')
        except Exception as e:
            self.logger.error(f"Error in building FAISS index: {e}")
            raise Exception(f'Error in building FAISS Index : {e}')


    
    def load_faiss_index(self):
        try:
            embeddings = HuggingFaceEmbeddings(
                model_name=self.emb_model_name,
                model_kwargs={'device': 'cpu'}
            )
            db = FAISS.load_local(self.faiss_index_path, embeddings)
            return db
        except Exception as e:
            self.logger.error(f"Error loading FAISS Index: {e}")
            raise Exception(f"Error loading FAISS Index: {e}")

    

    def setup_retriever(self, db, search_type_user=None, search_kwargs={'k': 10}):
        try:
            if search_type_user is not None:
                retriever = db.as_retriever(
                    search_type=search_type_user,
                    search_kwargs=search_kwargs
                )
            else:
                retriever = db.as_retriever(
                    search_kwargs=search_kwargs
                )
            return retriever
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Error setting up retriever: {e}")
            raise Exception(f"Error setting up retriever: {e}")

    def rerank_documents(self, documents, query, top_k=3):
        document_texts = [doc.page_content for doc in documents]
        pairs = [[query, text] for text in document_texts]
        scores = self.cross_encoder.predict(pairs)
        ranked_docs = sorted(zip(documents, scores), key=lambda x: x[1], reverse=True)
        return [doc for doc, score in ranked_docs[:top_k]]

    def retrieve_documents(self, db, prompt , search_type_user=None, search_kwargs={'k': 10}):
        try:
            self.qa_llm = self.setup_retriever(db, search_type_user=search_type_user, search_kwargs=search_kwargs)
            retrieved_docs = self.qa_llm.get_relevant_documents(prompt)
            top_docs = self.rerank_documents(retrieved_docs, prompt)
            return top_docs
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Error asking questions to LLM: {e}")
            raise Exception(f"Error asking LLM: {e}")



import asyncio
import os
from pathlib import Path
from querent.collectors.slack.slack_collector import SlackCollectorFactory
from querent.config.collector.collector_config import SlackCollectorConfig
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
import pytest
import uuid

from dotenv import load_dotenv

load_dotenv()


def get_collector_config():
    return SlackCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "channel_name": "C05TA5R7D88",
            "cursor": None,
            "include_all_metadata": 0,
            "inclusive": 0,
            "latest": 0,
            "limit": 100,
            "access_token": os.getenv("SLACK_ACCESS_KEY"),
            "name": "Slack-config",
            "config": {},
            "uri": "slack://"
        }
    )


@pytest.mark.asyncio
async def test_collect_and_ingest_generic_bytes():
    # Set up the collector
    collector_factory = SlackCollectorFactory()
    uri = Uri("slack://")
    config = get_collector_config()
    collector = collector_factory.resolve(uri, config)

    # Set up the ingestor
    ingestor_factory_manager = IngestorFactoryManager()
    ingestor_factory = await ingestor_factory_manager.get_factory("slack")
    ingestor = await ingestor_factory.create("slack", [])

    # Collect and ingest the PDF
    ingested_call = ingestor.ingest(collector.poll())

    async def poll_and_print():
        counter = 0
        async for ingested in ingested_call:
            assert ingested is not None
            if ingested is not "" or ingested is not None:
                counter += 1
        # Changing the number because messages before 90 days get deleted from slack
        assert counter > 0

    await poll_and_print()  # Notice the use of await here


if __name__ == "__main__":
    asyncio.run(test_collect_and_ingest_generic_bytes())



import asyncio
from pathlib import Path
import pytest
import os
import uuid
from dotenv import load_dotenv

from querent.config.collector.collector_config import (
    FSCollectorConfig,
    DriveCollectorConfig,
    SlackCollectorConfig,
)
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
from querent.collectors.collector_resolver import CollectorResolver
from querent.common.types.ingested_tokens import IngestedTokens


load_dotenv()


def drive_config():
    return DriveCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "drive_refresh_token": os.getenv("DRIVE_REFRESH_TOKEN"),
            "drive_token": os.getenv("DRIVE_TOKEN"),
            "drive_scopes": os.getenv("DRIVE_SCOPES"),
            "chunk_size": 1024 * 1024,
            "drive_client_id": os.getenv("DRIVE_CLIENT_ID"),
            "drive_client_secret": os.getenv("DRIVE_CLIENT_SECRET"),
            "specific_file_type": "application/pdf",
            "folder_to_crawl": "1BtLKXcYBrS16CX0R4V1X7Y4XyO9Ct7f8",
            "name": "Drive-config",
            "config": {},
            "uri": "drive://",
        }
    )


def slack_config():
    return SlackCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "channel_name": "C05TA5R7D88",
            "cursor": None,
            "include_all_metadata": 0,
            "inclusive": 0,
            "latest": 0,
            "limit": 100,
            "access_token": os.getenv("SLACK_ACCESS_KEY"),
            "name": "Slack-config",
            "config": {},
            "uri": "slack://",
        }
    )


@pytest.mark.asyncio
async def test_multiple_collectors_all_async():
    # Set up the collectors
    directories = ["./tests/data/pdf/", "./tests/data/pdf2/", "./tests/data/pdf3/"]
    collectors = [
        CollectorResolver().resolve(
            Uri("file://" + str(Path(directory).resolve())),
            FSCollectorConfig(
                config_source={
                    "id": str(uuid.uuid4()),
                    "root_path": directory,
                    "name": "Local-config",
                    "config": {},
                    "uri": "file://",
                }
            ),
        )
        for directory in directories
    ]

    collectors.append(
        CollectorResolver().resolve(
            Uri("drive://"),
            drive_config(),
        )
    )

    collectors.append(
        CollectorResolver().resolve(
            Uri("slack://"),
            slack_config(),
        )
    )

    for collector in collectors:
        await collector.connect()

    # Set up the result queue
    result_queue = asyncio.Queue()

    # Create the IngestorFactoryManager
    ingestor_factory_manager = IngestorFactoryManager(
        collectors=collectors, result_queue=result_queue
    )

    # Start the ingest_all_async in a separate task
    ingest_task = asyncio.create_task(ingestor_factory_manager.ingest_all_async())

    # Wait for the task to complete
    await asyncio.gather(ingest_task)

    # Optionally, check the result_queue for ingested data
    counter = 0
    unique_files = set()
    messages = 0
    while not result_queue.empty():
        ingested_data = await result_queue.get()
        if ingested_data is not None:
            if (
                isinstance(ingested_data, IngestedTokens)
                and ingested_data.is_token_stream
            ):
                messages += 1
            else:
                unique_files.add(ingested_data.file)
            counter += 1
    assert counter == 129
    assert len(unique_files) == 7
    assert messages > 0


if __name__ == "__main__":
    asyncio.run(test_multiple_collectors_all_async())



import spacy
from transformers import AutoTokenizer, AutoModelForTokenClassification
import torch
import nltk
from nltk.tokenize import sent_tokenize
from typing import Any, List, Tuple
from querent.logging.logger import setup_logger
import os
from querent.kg.ner_helperfunctions.dependency_parsing import Dependency_Parsing
from unidecode import unidecode


# nltk.data.path.append(
#     "/home/nishantg/querent/querent/nltk_resources/"
# )  ##need to configure still

"""
    Named Entity Recognition (NER) class for extracting entities and relationships from text.

    Attributes:
        ner_tokenizer (Tokenizer): Tokenizer for the NER model.
        ner_model (Model): Pre-trained NER model.
        logger (Logger): Logger instance for logging errors and information.

    Methods:
        tokenize_sentence(sentence: str) -> List[str]:
            Tokenizes a given sentence into individual tokens.

        get_chunks(tokens: List[str]) -> List[List[str]]:
            Splits tokens into smaller chunks for processing.

        extract_entities_from_chunk(chunk: List[str]) -> List[dict]:
            Extracts entities from a given chunk of tokens.

        combine_entities_wordpiece(entities: List[dict], tokens: List[str]) -> List[dict]:
            Combines entities that are split due to wordpiece tokenization.

        extract_binary_pairs(entities: List[dict], tokens: List[str], all_sentences: List[str], sentence_idx: int) -> List[Tuple[Tuple[dict, dict], dict]]:
            Extracts binary pairs of entities from the given entities list.

        extract_entities_from_sentence(sentence: str, sentence_idx: int, all_sentences: List[str]) -> Tuple[List[dict], List[Tuple[Tuple[dict, dict], dict]]]:
            Extracts entities and binary pairs from a given sentence.

        _token_distance(tokens: List[str], start_idx1: int, start_idx2: int, noun_chunk1: List[str], noun_chunk2: List[str]) -> int:
            Calculates the token distance between two entities.
    """


class NER_LLM:
    nlp = spacy.load('en_core_web_lg')
    def __init__(
        self, ner_model_name="dbmdz/bert-large-cased-finetuned-conll03-english",
        filler_tokens=None,
        provided_tokenizer=None,
        provided_model=None,
    ):  
        self.logger = setup_logger(__name__, "NER_LLM")
        if provided_tokenizer:
            self.ner_tokenizer = provided_tokenizer
        else:
            self.ner_tokenizer = AutoTokenizer.from_pretrained(ner_model_name)
        if provided_model:
            self.ner_model = provided_model
        else:
            self.ner_model = NER_LLM.load_model(ner_model_name, "NER")
        self.filler_tokens = filler_tokens or ["of", "a", "the", "in", "on", "at", "and", "or", "with","(",")","-"]
        
        

    def load_model(model_name, model_type):
        """Load the model, handling potential TensorFlow weights."""
        try:
            os.environ["CUDA_VISIBLE_DEVICES"] = "-1"  # Disable GPU if there's an issue
            return AutoModelForTokenClassification.from_pretrained(model_name, output_attentions=True)
        except OSError as e:
            os.environ["CUDA_VISIBLE_DEVICES"] = "-1"  # Disable GPU if there's an issue
            if "file named pytorch_model.bin" in str(e) and "file for TensorFlow weights" in str(e):
                return AutoModelForTokenClassification.from_pretrained(model_name, from_tf=True)
            raise OSError(
                f"Failed to load {model_type} model from {model_name}. Error: {e}"
            )
        except Exception as e:
            raise Exception(
                f"An unexpected error occurred while loading the model: {e}"
            )

    def validate(self) -> bool:
        return self.ner_model is not None and self.ner_tokenizer is not None

    @classmethod
    def get_class_variable(cls):
        return cls.nlp
    
    @staticmethod
    def split_into_sentences(document) -> List[str]:
        document = unidecode(document)
        sentences = []
        try:
            doc = NER_LLM.nlp(document)
            sentences = [sent.text for sent in doc.sents]
        except Exception as e:
            raise Exception(
                f"An unexpected error occurred while splitting the document into sentences: {e}"
            )
        return sentences

    def tokenize_sentence(self, sentence: str):
        return self.ner_tokenizer.tokenize(sentence)

    def _tokenize_and_chunk(self, data: str) -> List[Tuple[List[str], str, int]]:
        try:
            sentences = self.split_into_sentences(data)
            tokenized_sentences = []
            for idx, sentence in enumerate(sentences):
                sentence_tokens = self.tokenize_sentence(sentence)
                tokenized_sentences.append((sentence_tokens, sentence, idx))
        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to tokenize and chunk {e}")
            raise Exception(f"An error occurred while tokenizing: {e}")
        return tokenized_sentences

    def _token_distance(self, tokens, start_idx1, start_idx2, noun_chunk1, noun_chunk2):
        distance = 0
        for idx in range(start_idx1 + 1, start_idx2):
            token = tokens[idx]
            if (token not in self.filler_tokens and
                token not in noun_chunk1 and
                token not in noun_chunk2 and
                not token.startswith('##')):
                distance += 1
        return distance


    def transform_entity_pairs(self, entity_pairs):
        transformed_pairs = []
        sentence_group = {}
        for pair, metadata in entity_pairs:
            combined_sentence = ' '.join(filter(None, [
                metadata['previous_sentence'],
                metadata['current_sentence'],
                metadata['next_sentence']
            ]))
            if combined_sentence not in sentence_group:
                sentence_group[combined_sentence] = []
            sentence_group[combined_sentence].append(pair)

        for combined_sentence, pairs in sentence_group.items():
            for entity1, entity2 in pairs:
                meta_dict = {
                    "entity1_score": entity1['score'],
                    "entity2_score": entity2['score'],
                    "entity1_label": entity1['label'],
                    "entity2_label": entity2['label'],
                    "entity1_nn_chunk":entity1['noun_chunk'],
                    "entity2_nn_chunk":entity2['noun_chunk'],
                }
                new_pair = (entity1['entity'], combined_sentence, entity2['entity'], meta_dict)
                transformed_pairs.append(new_pair)

        return transformed_pairs

    def get_chunks(self, tokens: List[str], max_chunk_size=510):
        chunks = []
        k = 0
        while k < len(tokens):
            end = min(k + max_chunk_size, len(tokens))
            while end > k and end < len(tokens) and tokens[end].startswith("##"):
                end -= 1
            chunks.append(tokens[k:end])
            k = end
        return chunks

    def extract_entities_from_chunk(self, chunk: List[str]):
        results = []
        try:
            input_ids = self.ner_tokenizer.convert_tokens_to_ids(chunk)
            input_tensor = torch.tensor([input_ids])
            with torch.no_grad():
                outputs = self.ner_model(input_tensor)
            predictions = torch.argmax(outputs[0], dim=2)
            scores = torch.nn.functional.softmax(outputs[0], dim=2)
            label_ids = predictions[0].tolist()
            label_list = self.ner_model.config.id2label
            for idx, label_id in enumerate(label_ids):
                label = label_list[label_id]
                if label not in ["O", "[CLS]", "[SEP]", "[PAD]"]:
                    entity_info = {
                        "entity": chunk[idx],
                        "label": label,
                        "score": scores[0][idx][label_id].item(),
                        "start_idx": idx
                    }
                    results.append(entity_info)
        except Exception as e:
            self.logger.error(f"Error extracting entities from chunk: {e}")
            raise(f"Error extracting entities from chunk: {e}")
        return results

    def combine_entities_wordpiece(self, entities: List[dict], tokens: List[str]):
        combined_entities = []
        i = 0
        while i < len(entities):
            entity = entities[i]
            while i + 1 < len(entities) and entities[i + 1]["entity"].startswith("##"):
                entity["entity"] += entities[i + 1]["entity"][2:]
                entity["score"] = (entity["score"] + entities[i + 1]["score"]) / 2
                i += 1
            combined_entities.append(entity)
            i += 1
        final_entities = []
        for entity in combined_entities:
            entity_text = entity["entity"]
            start_idx = entity["start_idx"]
            while entity_text.startswith("##"):
                start_idx -= 1
                entity_text = tokens[start_idx] + entity_text[2:]
            entity["entity"] = entity_text
            final_entities.append(entity)

        return final_entities
    
    def combine_entities_byteencoding(self, entities: List[dict], tokens: List[str]):
        combined_entities = []
        i = 0
        while i < len(entities):
            entity = entities[i]
            while entity["start_idx"] > 0 and not tokens[entity["start_idx"]].startswith("▁"):
                entity["start_idx"] -= 1
                entity["entity"] = tokens[entity["start_idx"]] + entity["entity"]
            while i + 1 < len(entities) and entities[i + 1]["start_idx"] == entity["start_idx"] + 1 and not entities[i + 1]["entity"].startswith("▁"):
                entity["entity"] += entities[i + 1]["entity"]
                entity["score"] = (entity["score"] + entities[i + 1]["score"]) / 2
                i += 1
            combined_entities.append(entity)
            i += 1

        final_entities = []
        for entity in combined_entities:
            entity_text = entity["entity"]
            if entity_text.startswith("▁"):
                entity_text = entity_text[1:]
            entity["entity"] = entity_text
            final_entities.append(entity)

        return final_entities

    def extract_binary_pairs(self, entities: List[dict], tokens: List[str], all_sentences: List[str], sentence_idx: int):
        binary_pairs = []
        try:
            for i in range(len(entities)):
                for j in range(i + 1, len(entities)):
                    if entities[i]["start_idx"] + 1 == entities[j]["start_idx"]:
                        continue
                    distance = self._token_distance(tokens, entities[i]["start_idx"], entities[j]["start_idx"],entities[i]["noun_chunk"], entities[j]["noun_chunk"])
                    if distance <= 10:
                        pair = (entities[i], entities[j])
                        if pair not in binary_pairs:
                            metadata = {
                                "current_sentence": all_sentences[sentence_idx],
                                "previous_sentence": all_sentences[sentence_idx - 1] if sentence_idx > 0 else None,
                                "next_sentence": all_sentences[sentence_idx + 1] if sentence_idx < len(all_sentences) - 1 else None
                            }
                            binary_pairs.append((pair, metadata))
        except Exception as e:
            self.logger.error(f"Error extracting binary pairs: {e}")
            raise(f"Error extracting binary pairs: {e}")
        return binary_pairs
    
    def extract_fixed_entities_from_chunk(self, chunk: List[str], fixed_entities: List[str], entity_types: List[str], default_score=1.0):
        results = []
        try:
            for idx, token in enumerate(chunk):
                for entity_idx, entity in enumerate(fixed_entities):
                    if token.lower() == entity.lower():  # Case insensitive comparison
                        label = entity_types[entity_idx] if entity_idx < len(entity_types) else 'Unknown'
                        entity_info = {
                            "entity": token.lower(),
                            "label": label,
                            "score": default_score,
                            "start_idx": idx
                        }
                        results.append(entity_info)
        except Exception as e:
            self.logger.error(f"Error extracting fixed entities from chunk: {e}")
            raise Exception(f"Error extracting fixed entities from chunk: {e}")

        return results

    def extract_entities_from_sentence(self, sentence: str, sentence_idx: int, all_sentences: List[str], fixed_entities_flag: bool, fixed_entities: List[str],entity_types: List[str]):
        try:
            tokens = self.tokenize_sentence(sentence)
            chunks = self.get_chunks(tokens)
            all_entities = []
            for chunk in chunks:
                if fixed_entities_flag == False:
                    entities = self.extract_entities_from_chunk(chunk)
                else:
                    entities = self.extract_fixed_entities_from_chunk(chunk,fixed_entities, entity_types)
                all_entities.extend(entities)
            final_entities = self.combine_entities_wordpiece(all_entities, tokens)
            parsed_entities = Dependency_Parsing(entities=final_entities, sentence=sentence, model=NER_LLM.nlp)
            binary_pairs = self.extract_binary_pairs(parsed_entities.entities, tokens, all_sentences, sentence_idx)
            return parsed_entities.entities, binary_pairs
        except Exception as e:
            self.logger.error(f"Error extracting entities from sentence: {e}")
            raise(f"Error extracting entities from sentence: {e}")
    
    def remove_duplicates(self, data):
        seen = set()
        new_data = []

        for sublist in data:
            cleaned_sublist = []
            for sub_item in sublist:
                noun_chunk1 = sub_item[3]['entity1_nn_chunk']
                noun_chunk2 = sub_item[3]['entity2_nn_chunk']
                sentence = sub_item[1]
                if noun_chunk1 == noun_chunk2:
                    continue
                unique_key = (noun_chunk1, noun_chunk2, sentence)
                if unique_key not in seen:
                    seen.add(unique_key)
                    cleaned_sublist.append(sub_item)
            
            if cleaned_sublist:
                new_data.append(cleaned_sublist)

        return new_data





import pytest
from querent.callback.event_callback_interface import EventCallbackInterface
from querent.common.types.ingested_code import IngestedCode
from querent.common.types.ingested_images import IngestedImages
from querent.common.types.ingested_messages import IngestedMessages
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.querent_event import EventState, EventType
from querent.common.types.querent_queue import QuerentQueue
from querent.core.base_engine import BaseEngine
from querent.querent.querent import Querent
from querent.querent.resource_manager import ResourceManager

# Create input and output queues
input_queue = QuerentQueue()
resource_manager = ResourceManager()


# Define a simple mock LLM engine for testing
class MockLLMEngine(BaseEngine):
    def __init__(self, input_queue: QuerentQueue):
        super().__init__(input_queue)

    async def process_tokens(self, data: IngestedTokens):
        # await super().process_tokens(data)
        if data is None or data.is_error():
            # the LLM developer can raise an error here or do something else
            # the developers of Querent can customize the behavior of Querent
            # to handle the error in a way that is appropriate for the use case
            self.set_termination_event()
            return
        # Set the state of the LLM
        # At any given point during the execution of the LLM, the LLM developer
        # can set the state of the LLM using the set_state method
        # The state of the LLM is stored in the state attribute of the LLM
        # The state of the LLM is published to subscribers of the LLM
        current_state = EventState(EventType.Graph, 1.0, "anything", "dummy.txt")
        await self.set_state(new_state=current_state)

    async def process_code(self, data: IngestedCode):
        pass

    async def process_messages(self, data: IngestedMessages):
        return super().process_messages(data)

    async def process_images(self, data: IngestedImages):
        return super().process_images(data)

    def validate(self):
        return True


@pytest.mark.asyncio
async def test_querent_with_base_llm():
    # Put some input data into the input queue
    input_data = [
        IngestedTokens(file="", data="data1"),
        IngestedTokens(file="", data="data2"),
        IngestedTokens(file="", data="data3"),
        IngestedTokens(file="", data="", error="error"),
    ]
    for data in input_data:
        await input_queue.put(data)

    ### A Typical Use Case ###
    # Create an engine to harness the LLM
    llm_mocker = MockLLMEngine(input_queue)

    # Define a callback function to subscribe to state changes
    class StateChangeCallback(EventCallbackInterface):
        async def handle_event(self, event_type: EventType, event_state: EventState):
            print(f"New state: {event_state}")
            print(f"New state type: {event_type}")
            assert event_state.event_type == EventType.Graph

    # Subscribe to state change events
    # This pattern is ideal as we can expose multiple events for each use case of the LLM
    llm_mocker.subscribe(EventType.Graph, StateChangeCallback())

    ## one can also subscribe to other events, e.g. EventType.CHAT_COMPLETION ...

    # Create a Querent instance with a single MockLLM
    # here we see the simplicity of the Querent
    # massive complexity is hidden in the Querent,
    # while being highly configurable, extensible, and scalable
    # async architecture helps to scale to multiple querenters
    # How async architecture works:
    #   1. Querent starts a worker task for each querenter
    #   2. Querenter starts a worker task for each worker
    #   3. Each worker task runs in a loop, waiting for input data
    #   4. When input data is received, the worker task processes the data
    #   5. The worker task notifies subscribers of state changes
    #   6. The worker task repeats steps 3-5 until termination
    querent = Querent(
        [llm_mocker],
        resource_manager=resource_manager,
    )
    # Start the querent
    await querent.start()



from functools import lru_cache
import threading
from typing import Any
from rdflib import Graph
from rdflib.term import URIRef as RDFLibURIRef
from rdflib.term import BNode as RDFLibBNode
from rdflib.term import Literal as RDFLibLiteral
from querent.config.graph_config import GraphConfig
import time
from querent.graph.graph_namespace import NamespaceManager
from querent.graph.subject import Subject
from querent.graph.utils import URI, BNode, Literal
from rdflib.plugins.sparql import prepareQuery

from querent.logging.logger import setup_logger


class QuerentGraph(object):
    """
    A class for managing and manipulating an RDF graph with memory management features.
    """

    def __init__(
        self,
        graph_config: GraphConfig,
    ):
        """
        Initialize a QuerentGraph.

        Args:
            graph_config (GraphConfig): Configuration for the graph.
            memory_threshold_percent (float, optional): Memory threshold as a percentage of max_chunk_size. Default is None.
        """
        self.graph = Graph(
            identifier=RDFLibURIRef(graph_config.graph_identifier),
            store=graph_config.store,
        )

        self.memory_threshold = graph_config.memory_threshold
        self.format = graph_config.graph_format
        self.flush_on_serialize = graph_config.flush_on_serialize
        self.current_timestamp = time.time()
        self.logger = setup_logger(
            self.__class__.__name__, graph_config.graph_identifier
        )
        self._ns = NamespaceManager(self.graph)
        self.memory_cache = dict()
        self.lock = threading.Lock()

    def bind(self, prefix, namespace, override=True, replace=False):
        self._ns.bind(prefix, namespace, override, replace)

    def __add_subject(self, subjects, local_context=None):
        try:
            if not local_context:
                local_context = set()
            for t in subjects:
                s, p, o = t
                if isinstance(o, Subject) and o not in local_context:
                    local_context.add(o)
                    self.__add_subject(o, local_context)
                    o = o.subject
            # convert triple to RDFLib recognizable format
            triple = self._convert_to_rdflib((s, p, o))
            self.graph.add(triple)
        except Exception as e:
            self.logger.error(f"Failed to add subject: {subjects}")
            raise e

    def add_subjects(self, subjects):
        try:
            with self.lock:
                for s in subjects:
                    self.__add_subject(s)
        except Exception as e:
            self.logger.error(f"Failed to add subjects: {subjects}")
            raise e
        finally:
            self.graph.commit()

    def serialize(self):
        """
        Serialize the graph.

        Returns:
            str: Serialized RDF data.
        """
        try:
            if self.format.lower() in ("ttl", "turtle"):
                b_string = self.graph.serialize(
                    format=self.format,
                    namespace_manager=self._ns,
                )
            elif self.format.lower() == "json-ld":
                b_string = self.graph.serialize(
                    format=self.format,
                    contexts=self._ns,
                )
            else:
                b_string = self.graph.serialize(
                    format=self.format,
                )
            if isinstance(b_string, bytes):
                b_string = b_string.decode("UTF-8")

            if self.flush_on_serialize:
                self.graph.remove((None, None, None))

            return b_string
        except Exception as e:
            self.logger.error("Failed to serialize graph")
            self.logger.error(e)
            raise e

    def parse(self, content):
        try:
            self.graph.parse(data=content, format=self.format)
        except Exception as e:
            raise e

    def _is_rdf_type(self, uri: Any) -> bool:
        try:
            return (
                isinstance(self._resolve_uri(uri), RDFLibURIRef)
                or isinstance(self._resolve_uri(uri), RDFLibBNode)
                or isinstance(self._resolve_uri(uri), RDFLibLiteral)
            )
        except Exception as e:
            self.logger.error(f"Failed to resolve URI: {uri.value}")
            return False

    @lru_cache()
    def _resolve_uri(self, uri: URI) -> RDFLibURIRef:
        """
        Convert a URI object into a RDFLib URIRef, including resolve its context

        :param uri: URI
        :return: rdflib.URIRef
        """
        try:
            return self._ns.parse_uri(uri.value)
        except Exception as e:
            self.logger.error(f"Failed to resolve URI: {uri.value}")
            raise e

    def _convert_to_rdflib(self, triple):
        """
        Convert a Node Quad to RDFLib Quad
        """
        s, p, o = triple
        sub = self._resolve_uri(s) if isinstance(s, URI) else RDFLibBNode(s.value)
        pred = self._resolve_uri(p)
        if isinstance(o, URI):
            obj = self._resolve_uri(o)
        elif isinstance(o, Subject):
            if isinstance(o.subject, URI):
                obj = self._resolve_uri(o.subject)
            else:
                obj = RDFLibBNode(o.subject.value)
        elif isinstance(o, BNode):
            obj = RDFLibBNode(o.value)
        elif isinstance(o, Literal):
            obj = RDFLibLiteral(o.value, o.lang, o.raw_type)
        else:
            raise Exception("Object must be URI, BNode or Literal")

        return (sub, pred, obj)

    def clear(self):
        self.graph.remove((None, None, None))

    @property
    def get_current_memory_usage(self):
        # Calculate the current memory usage
        return self.calculate_memory_usage()

    @lru_cache()
    def calculate_memory_usage(self):
        # Calculate and store the memory usage
        memory_usage = self.calculate_actual_memory_usage()
        self.memory_cache["memory_usage"] = memory_usage
        return memory_usage

    def calculate_actual_memory_usage(self):
        # Calculate the actual memory usage
        return self.graph.serialize(format="nt").__sizeof__()

    def execute_query(self, query_str: str):
        """
        Execute a query. Example usage to get all subjects with specific predicate and object

        graph = QuerentGraph()
        Add few triples and then

        sparql_query = \"""
        SELECT ?subject
        WHERE {
            ?subject <http://example.org/predicate> "Example Object" .
        }
        \"""
        results = querent_graph.execute_query(sparql_query)
        print(results)
        """
        try:
            query = prepareQuery(query_str)
            return list(self.graph.query(query))
        except Exception as e:
            self.logger.error(f"Failed to execute query: {query_str}")
            raise e



# "TESTS PASSED, NEEDS KEYS"

# import asyncio

# from querent.config.collector_config import AzureCollectConfig
# from querent.collectors.collector_resolver import CollectorResolver
# from querent.collectors.azure.azure_collector import AzureCollectorFactory
# from querent.common.uri import Uri
# from querent.config.collector_config import CollectorBackend
# import pytest
# import os
# from dotenv import load_dotenv

# load_dotenv()

# azure_connection_string = os.getenv("AZURE_STORAGE_CONNECTION_STRING")
# azure_account_url = os.getenv("AZURE_STORAGE_ACCOUNT_URL")
# azure_account_key = os.getenv("AZURE_STORAGE_ACCOUNT_KEY")


# @pytest.fixture
# def azure_config():
#     config = AzureCollectConfig(
#         connection_string="",import asyncio

# from querent.config.collector_config import AzureCollectConfig
# from querent.collectors.collector_resolver import CollectorResolver
# from querent.collectors.azure.azure_collector import AzureCollectorFactory
# from querent.common.uri import Uri
# from querent.config.collector_config import CollectorBackend
# import pytest
# import os
# from dotenv import load_dotenv

# load_dotenv()

# azure_connection_string = os.getenv("AZURE_STORAGE_CONNECTION_STRING")
# azure_account_url = os.getenv("AZURE_STORAGE_ACCOUNT_URL")
# azure_account_key = os.getenv("AZURE_STORAGE_ACCOUNT_KEY")


# @pytest.fixture
# def azure_config():
#     config = AzureCollectConfig(
#         connection_string="",
#         account_url=azure_account_url,
#         credentials=azure_account_key,
#         chunk=1024,
#         container="testfiles",
#         prefix="",
#     )
#     return config


# def test_azure_collector_factory():
#     factory = AzureCollectorFactory()
#     assert factory.backend() == CollectorBackend.AzureBlobStorage


# # Modify this function to test the Azure collector


# @pytest.mark.asyncio
# async def test_azure_collector(azure_config):
#     config = azure_config
#     uri = Uri("azure://" + config.container)
#     resolver = CollectorResolver()
#     collector = resolver.resolve(uri, config)
#     assert collector is not None

#     async def poll_and_print():
#         counter = 0
#         async for result in collector.poll():
#             assert not result.is_error()
#             chunk = result.unwrap()
#
#             if chunk:
#                 counter += 1
#         assert counter == 1433

#     await poll_and_print()


# if __name__ == "__main__":
#     asyncio.run(test_azure_collector())

#         account_url=azure_account_url,
#         credentials=azure_account_key,
#         chunk=1024,
#         container="testfiles",
#         prefix="",
#     )
#     return config


# def test_azure_collector_factory():
#     factory = AzureCollectorFactory()
#     assert factory.backend() == CollectorBackend.AzureBlobStorage


# # Modify this function to test the Azure collector


# @pytest.mark.asyncio
# async def test_azure_collector(azure_config):
#     config = azure_config
#     uri = Uri("azure://" + config.container)
#     resolver = CollectorResolver()
#     collector = resolver.resolve(uri, config)
#     assert collector is not None

#     async def poll_and_print():
#         counter = 0
#         async for result in collector.poll():
#             assert not result.is_error()
#             chunk = result.unwrap()
#
#             if chunk:
#                 counter += 1
#         assert counter == 1433

#     await poll_and_print()


# if __name__ == "__main__":
#     asyncio.run(test_azure_collector())



from typing import Optional
from pydantic import BaseModel
import os

from querent.common.types.resource_config_keys import ResourceConfigKey


class ResourceConfig(BaseModel):
    id: str
    max_workers_allowed: Optional[int]
    max_workers_per_collector: Optional[int]
    max_workers_per_engine: Optional[int]
    max_workers_per_querent: Optional[int]

    def __init__(self, config_source=None, **kwargs):
        # if kwargs:
        #     raise ValueError(
        #         "Config values must be provided within a dictionary via 'config_source' parameter."
        #     )

        if config_source:
            config_data = self.load_config(config_source)
            for config_key in ResourceConfigKey:
                key = config_key.value
                if key in config_data:
                    setattr(self, key, config_data[key])

        # else:
        #     raise ValueError("Please pass config")

    @classmethod
    def load_config(cls, config_source) -> dict:
        if isinstance(config_source, dict):
            # If config source is a dictionary, return a dictionary
            cls.config_data = config_source
        else:
            raise ValueError("Invalid config. Must be a valid dictionary")

        env_vars = dict(os.environ)
        cls.config_data.update(env_vars)
        return cls.config_data

    @classmethod
    def get_full_config(cls):
        return cls.config_data

    @classmethod
    def get(cls, key: ResourceConfigKey, default=None):
        """
        Get a specific configuration value by key.
        Args:
            key (ConfigKey): The key for the configuration value.
            default: The default value to return if the key is not found.
        Returns:
            The configuration value if found, otherwise the default value.
        """
        return cls.config_data.get(key, default)



import asyncio
import pytest
import os
from querent.collectors.collector_resolver import CollectorResolver
from querent.config.collector.collector_config import JiraCollectorConfig
from querent.common.uri import Uri
import uuid
from dotenv import load_dotenv

from querent.ingestors.ingestor_manager import IngestorFactoryManager

load_dotenv()


@pytest.fixture
def jira_config():
    return JiraCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "jira_server": "https://querent.atlassian.net/",
            "jira_username": "puneet@querent.xyz",
            "jira_api_token": os.getenv("JIRA_API_TOKEN"),
            "jira_project": "Querent1",
            "jira_query": "project=Querent1",
            "name": "Jira-config",
            "config": {},
            "uri": "jira://",
        }
    )


@pytest.mark.asyncio
async def test_jira_ingestor(jira_config):
    uri = Uri("jira://")
    resolver = CollectorResolver()
    collector = resolver.resolve(uri, jira_config)
    assert collector is not None
    await collector.connect()

    # Set up the ingestor
    ingestor_factory_manager = IngestorFactoryManager()
    ingestor_factory = await ingestor_factory_manager.get_factory("jira")
    ingestor = await ingestor_factory.create("jira", [])
    ingested_call = ingestor.ingest(collector.poll())

    async def poll_and_print():
        counter = 0
        async for ingested in ingested_call:
            assert ingested is not None
            if ingested is not "" or ingested is not None:
                counter += 1
        # 1 extra ingestedToken signifying end of file
        assert counter == 2

    await poll_and_print()  # Notice the use of await here


if __name__ == "__main__":
    asyncio.run(test_jira_ingestor())



import asyncio
import pytest
import os
from querent.collectors.collector_resolver import CollectorResolver
from querent.config.collector.collector_config import (
    CollectorBackend,
    EmailCollectorConfig,
)
from querent.common.uri import Uri
import uuid
from dotenv import load_dotenv

from querent.ingestors.ingestor_manager import IngestorFactoryManager

load_dotenv()


@pytest.fixture
def email_config():
    return EmailCollectorConfig(
        config_source={
            "backend": "email",
            "id": str(uuid.uuid4()),
            "imap_server": "imap.gmail.com",
            "imap_port": 993,
            "imap_username": "puneet@querent.xyz",
            "imap_password": os.getenv("IMAP_PASSWORD"),
            "imap_folder": "[Gmail]/Drafts",
            "imap_certfile": None,
            "imap_keyfile": None,
            "name": "Email-config",
            "config": {},
            "uri": "email://",
        }
    )


@pytest.mark.asyncio
async def test_email_ingestor(email_config):
    uri = Uri("email://")
    resolver = CollectorResolver()
    collector = resolver.resolve(uri, email_config)
    assert collector is not None
    await collector.connect()
    # Set up the ingestor
    ingestor_factory_manager = IngestorFactoryManager()
    ingestor_factory = await ingestor_factory_manager.get_factory("email")
    ingestor = await ingestor_factory.create("email", [])
    ingested_call = ingestor.ingest(collector.poll())

    async def poll_and_print():
        counter = 0
        async for ingested in ingested_call:
            assert ingested is not None
            if ingested != "" or ingested is not None:
                counter += 1
        assert counter == 4

    await poll_and_print()  # Notice the use of await here


if __name__ == "__main__":
    asyncio.run(test_email_ingestor())



import asyncio
from pathlib import Path
from querent.collectors.fs.fs_collector import FSCollectorFactory
from querent.config.collector.collector_config import FSCollectorConfig
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
import pytest
import uuid

count = 0


@pytest.mark.asyncio
async def test_collect_and_ingest_pdf():
    # Set up the collector
    output_queue = asyncio.Queue()
    logger_queue = asyncio.Queue()
    stop_event = asyncio.Event()
    directories = ["./tests/data/pdf/", "./tests/data/pdf2/", "./tests/data/pdf3/"]
    workers = [
        create_workers(i, output_queue, logger_queue, stop_event) for i in directories
    ]

    # stop_event_tasks = asyncio.create_task(stop_workers(logger_queue, stop_event))
    queue_tasks = asyncio.create_task(see_contents(output_queue))
    asyncio.gather(*workers)
    await output_queue.join()
    await queue_tasks


async def see_contents(output_queue):
    while True:
        data = await output_queue.get()
        if data == "DONE":
            global count
            count = count + 1
        if count == 3:
            break
        else:
            output_queue.task_done()
            print(data)


async def create_workers(directory, output_queue, logger_queue, stop_event):
    try:
        collector_factory = FSCollectorFactory()
        uri = Uri("file://" + str(Path(directory).resolve()))
        config = FSCollectorConfig(
            config_source={
                "id": str(uuid.uuid4()),
                "root_path": directory,
                "name": "Local-config",
                "config": {},
                "uri": "file://",
            }
        )
        collector = collector_factory.resolve(uri, config)
        ingestor_factory_manager = IngestorFactoryManager()
        ingestor_factory = await ingestor_factory_manager.get_factory("pdf")
        ingestor = await ingestor_factory.create("pdf", [])
        ingested_call = ingestor.ingest(collector.poll())
        async for ingested in ingested_call:
            if ingested.data is None:
                # continue
                await logger_queue.put("1 file received")
                print("Putting 1 file in logger queue")
            else:
                await output_queue.put(ingested.data)
    except Exception as e:
        print("Got an exception:  ", e)

    finally:
        # Always ensure we send the "DONE" signal, even if an exception occurred
        await output_queue.put("DONE")


if __name__ == "__main__":
    asyncio.run(test_collect_and_ingest_pdf())



# from querent.kg.rel_helperfunctions.fixed_relationships import FixedRelationshipExtractor  # Adjust the import according to your project structure

# def test_fixed_entity_extractor():
#     full_document_text = """1. Introduction
# The impact of climate change on glaciers of the world has been a topic of active concern
# in recent times (Syed et al. 2007; Bagla 2009; Cogley et al. 2010). While glacial melt is a
# major source of freshwater that contributes to global mean sea level rise (Meier et al. 2007;
# Jacob et al. 2012), decreasing glacial mass has an impact on the freshwater resources of
# some of the largest river basins in the world, including that of the Ganges (e.g. Kumar,
# Singh, and Sharma 2005; Mall et al. 2006; Immerzeel, van Beek, and Bierkens 2010;
# Kaser, Grosshauser, and Marzeion 2010). Fluctuations in the recession rate of glaciers dur-
# ing recent years have initiated widespread discussions, especially in the context of global
# warming and its effects (Dyurgerov and Meier 2005). Similar ice mass changes have also
# been reported in the Garhwal region of the Himalayas (Kulkarni et al. 2005; Bhambri,
# Bolch, and Chaujar 2011a, 2011b; Kulkarni et al. 2011; Bolch et al. 2012). However, the
# *Corresponding author. Email: tsyed.ismu@gmail.com
# © 2013 Taylor & FrancisDownloaded by [Tajdarul Syed] at 06:13 15 October 2013
# 8654
# P. Saraswat et al.
# status of the Gangotri glacier warrants further investigation, particularly in the context of
# climate change resulting in continuous retreat, negative mass balance, and early melting of
# seasonal snow cover (Negi et al. 2012).
# Glaciers cover about 40,800 km2 of the Himalayan and Karakoram mountain region,
# and are one of the main sources of water to some of the largest rivers in the world, including
# the Ganges, Brahmaputra, and Indus (Kulkarni et al. 2011; Bajracharya and Shrestha 2011;
# Bolch et al. 2012). Approximately 10% of the summer discharge of the Ganges is attributed
# to melt waters from glaciers (Barnett, Adam, and Letternmaier 2005; Jain, Agarwal, and
# Singh 2007; Immerzeel, van Beek, and Bierkens 2010; Bolch et al. 2012). Rapid depletion
# of glaciers has had adverse effects on the flow regime of major Himalayan rivers, and can
# even lead to catastrophic events such as glacial lake outbursts (Govindha 2010; Shrestha,
# Eriksson, and Mool 2010; Ashraf, Naz, and Roohi 2012) and consequent flooding in the
# upper reaches of these rivers, affecting the lives of millions of people residing in the Indo-
# Gangetic plains (Richardson and Reynolds 2000; Bolch et al. 2012).
# In spite of their hydrologic, climatic, and socioeconomic significance, opportunities for
# precise and continuous monitoring of the Himalayan glaciers are limited by logistical and
# terrain-induced difficulties. Observations of Earth’s surface using satellite data have proven
# to be very useful in such monitoring; recent developments in high-resolution image acqui-
# sition have facilitated more precise monitoring of glacier movement (Luckman, Quincey,
# and Bevan 2007; Kumar, Venkataraman, and Hogda 2011; Kumar et al. 2011). Further,
# satellite data enables a gross analysis of glacier mass budgets, overcoming some of the
# problems of accessibility and sustainability of long-term measurements, that can later be
# verified with ground-based surveys (Kumar et al. 2008). The potential of remote sensing
# for glacier mass balance and velocity mapping has been demonstrated with optical (e.g.
# Kääb 2005; Scherler, Leprince, and Strecker 2008), synthetic aperture radar (SAR) (e.g.
# Luckman, Quincey, and Bevan 2007), and thermal infrared sensors (e.g. Nakawo, Yabuki,
# and Sakai 1999), and elevation models (Bolch, Pieczonka, and Benn 2011), among others.
# Developments in deriving flow rates and monitoring Himalayan glacier retreat using optical
# images have been made by Kääb (2005). However, this approach is sometimes limited by
# weather, clouds, and shadows in areas of high relief.
# In a novel approach, the current study presents one of the most comprehensive assess-
# ments of the Gangotri glacier in recent times (2004–2011). The methodology entails the
# utilization of interferometric SAR (InSAR) coherence and sub-pixel offset tracking. While
# complementing most previous studies, the result presented here establishes the effective-
# ness of the techniques implemented to produce robust estimates of areal changes and glacier
# surface velocity in near real time. But, perhaps most importantly, this is one of the few stud-
# ies which has shown the melting trend of Gangotri glacier over a considerably continuous
# period during recent times (2004–2011).
# 2. Study area
# The Gangotri glacier is a valley-type glacier and one of the largest Himalayan glaciers
# located in Uttarkashi district of Uttarakhand, India (Figure 1). Extending between the lat-
# itudes 30◦ 43 22 N–30◦ 55 49 N and longitudes 79◦ 4 41 E–79◦ 16 34 E, Gangotri
# is the only major Himalayan glacier that flows towards the northwest. It spans a length
# of 30.2 km, its width varies between 0.20 and 2.35 km, and it thereby covers an area of
# about 86.32 km2 . While the average thickness of the Gangotri glacier is ∼200 m, its sur-
# face elevation varies from 4000 to 7000 m above mean sea level (Jain 2008). Gangotri has
# three main tributaries, namely the Raktvarna, the Chaturangi, and the Kirti, with lengthsInternational Journal of Remote Sensing
# 70° 0′ 00″ E 75° 0′ 00″ E 80° 0′ 00″ E 85° 0′ 00″ E 90° 0′ 00″ E
# 79º 5' E
# 440 km
# 79º 10' E
# Raktvarna
# 30° 55′ N
# 220
# 30° 55′ N
# 35° 0′ 00″ N
# 0
# 8655
# Chaturangi
# 30° 0′ 00″ N
# 30° 50′ N
# tri
# go
# an
# G
# 25° 0′ 00″ N
# 30° 50′ N
# Kirti
# Shivling
# Hills
# Bhagirathi
# Hills
# 20° 0′ 00″ N
# 79° 5′ E
# Downloaded by [Tajdarul Syed] at 06:13 15 October 2013
# 0
# 79° 10′ E
# 2
# 4
# 6
# km
# Figure 1. Location of the Gangotri glacier in the overall Indian perspective. Also shown is a mag-
# nified view of the study area using an ASTER image illustrating the locations of the Gangotri glacier
# and its tributaries (Kirti, Raktvarna, and Chaturangi). The figure also marks the locations of the
# Shivling and Bhagirathi hills.
# of 15.90, 22.45, and 11.05 km, respectively (Figure 1). Additionally, there are more than
# 18 small tributary glaciers of Gangotri and its tributaries. Gangotri and other glaciers in
# this region are mostly fed by the summer monsoon and partly by winter snow. Western
# disturbances cause heavy snowfall from December to March over this region (Thayyen and
# Gergan 2010). Generally, seasonal melting star
# """  

#     fixed_relationships = ["impact of", "resulting in"]  # Example relationships

#     extractor = FixedRelationshipExtractor(fixed_relationships)
#     result = extractor.find_relationship_sentences(full_document_text)
#     reduction_percentage = extractor.measure_reduction(full_document_text, result)

#     assert "impact of" in result
#     assert "resulting in" in result



from enum import Enum


class IngestorError(Enum):
    EOF = "End of File"
    ETCD = "ETCD Error"
    NETWORK = "Network Error"
    TIMEOUT = "Timeout"
    UNKNOWN = "Unknown Error"
    FILE_NOT_FOUND = "File Not Found"
    IOERROR = "IOError"
    UIE = "UnidentifiedImageError"
    WRONGPPTFILE = "Wrong PPt file"
    INVALIDXMLERROR = "Invalid Xml Error"
    BADZIPFILE = "BadZipFile"
    UNICODEDECODEERROR = "UnicodeDecodeError"
    LOOKUPERROR = "LookupError"
    TYPEERROR = "TypeError"
    UNKNOWNVALUEERROR = "UnknownValueError"
    REQUESTERROR = "RequestError"
    INDEXERROR = "IndexError"
    CSVERROR = "CsvError"
    RUNTIMEERROR = "RuntimeError"
    JSONDECODEERROR = "JsonDecodeError"
    DOCUMENTERROR = "DocumentError"
    SHELLERROR = "ShellError"
    PERMISSIONERROR = "PermissionError"
    OSERROR = "OSError"
    CONNECTIONERROR = "ConnectionError"
    SLACKERROR = "SlackAPIError"
    POLLINGERROR = "PollingError"


class IngestorErrorBase(Exception):
    def __init__(self, error_code, message=None) -> None:
        super().__init__(message)
        self.error_code = error_code


class FileNotFoundError(IngestorErrorBase):
    """Error function for file not found"""

    def __init__(self, message=None):
        super().__init__(message=message, error_code=IngestorError.FILE_NOT_FOUND)


class IOError(IngestorErrorBase):
    """Error function for I/O Errors"""

    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.IOERROR, message)


class UnidentifiedImageError(IngestorErrorBase):
    """Error function for UnidentifiedImageError"""

    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.UIE, message)


class WrongPptFileError(IngestorErrorBase):
    """Error function if ppt file is not being passed into ppt ingestor"""

    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.WRONGPPTFILE, message)


class InvalidXmlError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.INVALIDXMLERROR, message)


class BadZipFile(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.BADZIPFILE, message)


class UnicodeDecodeError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.UNICODEDECODEERROR, message)


class LookupError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.LOOKUPERROR, message)


class TypeError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.TYPEERROR, message)


class UnknownValueError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.UNKNOWNVALUEERROR, message)


class RequestError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.REQUESTERROR, message)


class IndexErrorException(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.INDEXERROR, message)


class UnknownError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.UNKNOWN, message)


class CsvError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.CSVERROR, message)


class RuntimeError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.RUNTIMEERROR, message)


class JsonDecodeError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.JSONDECODEERROR, message)


class DocumentError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.DOCUMENTERROR, message)


class ShellError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.SHELLERROR, message)


class PermissionError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.PERMISSIONERROR, message)


class OSError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.OSERROR, message)


class ConnectionError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.CONNECTIONERROR, message)


class SlackApiError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.SLACKERROR, message)


class PollingError(IngestorErrorBase):
    def __init__(self, message=None) -> None:
        super().__init__(IngestorError.POLLINGERROR, message)



# import asyncio
# import json
# import pytest
# from querent.callback.event_callback_interface import EventCallbackInterface
# from querent.common.types.querent_queue import QuerentQueue
# from querent.common.types.querent_event import EventState, EventType
# from querent.common.types.ingested_tokens import IngestedTokens
# from querent.common.types.querent_queue import QuerentQueue
# from querent.config.core.bert_llm_config import LLM_Config
# from querent.core.transformers.bert_llm import BERTLLM
# from querent.querent.resource_manager import ResourceManager
# from querent.querent.querent import Querent


# @pytest.mark.asyncio
# @pytest.mark.parametrize("input_data,ner_model_name, llm_class,expected_entities,filter_entities", [
#     #("Nishant is working from Delhi. Ansh is working from Punjab. Ayush is working from Odisha. India is very good at playing cricket. Nishant is working from Houston.", "dbmdz/bert-large-cased-finetuned-conll03-english", BERTLLM, ["http://geodata.org/Nishant", "http://geodata.org/Delhi"], False),
#     #("Nishant is working from Delhi. Ansh is working from Punjab. Ayush is working from Odisha. India is very good at playing cricket. Nishant is working from Houston.", "dbmdz/bert-large-cased-finetuned-conll03-english", BERTLLM, ["http://geodata.org/Nishant", "http://geodata.org/Delhi"], False),
# #     ("""In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM)
# # record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM)
# # using organic carbon stable isotopes and biostratigraphic constraints. We suggest that
# # climate and tectonic perturbations in the upstream North American catchments can induce
# # a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately
# # in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom-
# # modation and deposition of a shale interval when coarse-grained terrigenous material
# # was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi-
# # ment supply during the PETM, which is archived as a particularly thick sedimentary
# # section in  the deep-sea fans of the GoM basin. The Paleocene–Eocene Thermal Maximum (PETM) (ca. 56 Ma) was a rapid global warming event characterized by the rise of temperatures to5–9 °C (Kennett and Stott, 1991), which caused substantial environmental changes around the globe.""",
# # "botryan96/GeoBERT", BERTLLM, ["tectonic perturbations","downstream sectors"], True)])

#     ("""Beyond methane: Towards a theory for the Paleocene–Eocene Thermal Maximum John A. Higgins⁎, Daniel P. Schrag Department of 
#      Earth and Planetary Sciences, Harvard University, 20 Oxford St., Cambridge, MA, 02138, United States Received 15 September 2005; 
#      received in revised form 24 February 2006; accepted 1 March 2006 Available online 6 May 2006 Editor: H. Elderfield Abstract 
#      Extreme global warmth and an abrupt negative carbon isotope excursion during the Paleocene–Eocene Thermal Maximum (PETM) 
#      have been attributed to a massive release of methane hydrate from sediments on the continental slope [G.R. Dickens, 
#      J.R.O'Neil, D.K. Rea, R.M. Owen, Dissociation of oceanic methane hydrate as a cause of the carbon isotope excursion at 
#      the end of the Paleocene, Paleoceanography 10 (1995) 965 –971.]. However, the magnitude of the warming (5 to 6 °C 
#      [J.C. Zachos, M.W. Wara, S. Bohaty, M.L. Delaney, M.R. Petrizzo, A. Brill, T.J. Bralower, I. Premoli-Silva, 
#      A transient rise in tropical sea surfacetemperature during the Paleocene–Eocene Thermal Maximum, Science 302 (2003) 
#      1551 –1554.,J.P. Kennett, L.D. Stott, Abrupt deep-sea warming, paleoceanographic changes and benthic extinctions at the 
#      end of the Paleocene, Nature 353 (1991) 225 –228.]) and rise in the depth of the CCD ( N2 km; [J.C. Zachos, U. Rohl, S.A. 
#      Schellenberg, D. Hodell, E. Thomas, A. Sluijs, C. Kelly, H. McCarren, D. Kroon, I. Raffi, L.J. Lourens, M. Nicolo, Rapid 
#      acidification of the ocean during the Paleocene–Eocene Thermal Maximum, Science 308 (2005) 1611 –1615.]) indicate that 
#      the size of the carbon addition was larger than can be accounted for by the methane hydrate hypothesis. 
#      Additional carbon sources associated with methane hydrate release (e.g. pore-water venting andturbidite oxidation) are also 
#      insufficient. We find that the oxidation of at least 5000 Gt C of organic carbon is the most likelyexplanation for the 
#      observed geochemical and climatic changes during the PETM, for which there are several potential mechanisms. 
#      Production of thermogenic CH 4and CO 2during contact metamorphism associated with the intrusion of a large igneous province 
#      into organic rich sediments [H. Svensen, S. Planke, A. Malthe-Sorenssen, B. Jamtveit, R. Myklebust, T.R. Eidem,S.S. Rey, 
#      Release of methane from a volcanic basin as a mechanism for initial Eocene global warming, Nature 429 (2004).] 
#      is capable of supplying large amounts of carbon, but is inconsistent with the lack of extensive carbon loss in 
#      metamorphosed sediments, as well as the abrupt onset and termination of carbon release during the PETM.""",
# "botryan96/GeoBERT", BERTLLM, ["tectonic perturbations","downstream sectors"], True)])

# async def test_bertllm_ner_tokenization_and_entity_extraction(input_data, ner_model_name, llm_class, expected_entities, filter_entities):
#     input_queue = QuerentQueue()
#     resource_manager = ResourceManager()
#     ingested_data = IngestedTokens(file="dummy_1_file.txt", data=input_data)
#     await input_queue.put(ingested_data)
#     ingested_data = IngestedTokens(file="dummy_1_file.txt", data=None)
#     await input_queue.put(ingested_data)
#     await input_queue.put(IngestedTokens(file="dummy_2_file.txt", data=None, error="error"))
#     bert_llm_config = LLM_Config(
#         ner_model_name=ner_model_name,
#         enable_filtering=filter_entities,
#         filter_params={
#             'score_threshold': 0.5,
#             'attention_score_threshold': 0.1,
#             'similarity_threshold': 0.5,
#             'min_cluster_size': 5,
#             'min_samples': 3,
#             'cluster_persistence_threshold':0.1
#         }
#             # ,fixed_entities = ['carbon isotope']
#             # , sample_entities=['B-GeoPetro']
#     )
#     llm_instance = llm_class(input_queue, bert_llm_config)
#     class StateChangeCallback(EventCallbackInterface):
#         async def handle_event(self, event_type: EventType, event_state: EventState):
#             assert event_state.event_type == EventType.Graph
#             triple = json.loads(event_state.payload)
#             print("triple: {}".format(triple))
#             assert isinstance(triple['subject'], str) and triple['subject']
#     llm_instance.subscribe(EventType.Graph, StateChangeCallback())
#     querent = Querent(
#         [llm_instance],
#         resource_manager=resource_manager,
#     )

#     await querent.start()




from io import BytesIO
from PyPDF2 import PdfReader
import requests
import re
from requests.exceptions import RequestException
from bs4 import BeautifulSoup
from newspaper import Article, ArticleException, Config
from requests_html import HTMLSession
import random
from lxml import html

from querent.logging.logger import setup_logger

logger = setup_logger(__name__, "webpage_extractor.log")

USER_AGENTS = [
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:66.0) Gecko/20100101 Firefox/66.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_5) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.1.1 Safari/605.1.15",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/64.0.3282.140 Safari/537.36 Edge/17.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:72.0) Gecko/20100101 Firefox/72.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/78.0.3904.97 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.3 Safari/605.1.15",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:80.0) Gecko/20100101 Firefox/80.0",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.87 Safari/537.36",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 13_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.4 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:73.0) Gecko/20100101 Firefox/73.0",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.129 Safari/537.36"
]

class WebpageExtractor:

    def __init__(self, num_extracts=3):
        """
        Initialize the WebpageExtractor class.
        """
        self.num_extracts = num_extracts

    def extract_with_3k(self, url):
        """
        Extract the text from a webpage using the 3k method.

        Args:
            url (str): The URL of the webpage to extract from.

        Returns:
            str: The extracted text.
        """
        try:
            if url.lower().endswith(".pdf"):
                response = requests.get(url)
                response.raise_for_status()

                with BytesIO(response.content) as pdf_data:
                    reader = PdfReader(pdf_data)
                    content = " ".join([reader.getPage(i).extract_text() for i in range(reader.getNumPages())])

            else:
                config = Config()
                config.browser_user_agent = random.choice(USER_AGENTS)
                config.request_timeout = 10
                session = HTMLSession()

                response = session.get(url)
                response.html.render(timeout=config.request_timeout)
                html_content = response.html.html

                article = Article(url, config=config)
                article.set_html(html_content)
                article.parse()
                content = article.text.replace('\t', ' ').replace('\n', ' ').strip()

            return content[:1500]

        except ArticleException as ae:
            logger.error(f"Error while extracting text from HTML (newspaper3k): {str(ae)}")
            return f"Error while extracting text from HTML (newspaper3k): {str(ae)}"

        except RequestException as re:
            logger.error(f"Error while making the request to the URL (newspaper3k): {str(re)}")
            return f"Error while making the request to the URL (newspaper3k): {str(re)}"

        except Exception as e:
            logger.error(f"Unknown error while extracting text from HTML (newspaper3k): {str(e)}")
            return ""

    def extract_with_bs4(self, url):
        """
        Extract the text from a webpage using the BeautifulSoup4 method.

        Args:
            url (str): The URL of the webpage to extract from.

        Returns:
            str: The extracted text.
        """
        headers = {
            "User-Agent": random.choice(USER_AGENTS)
        }

        try:
            response = requests.get(url, headers=headers, timeout=10)
            if response.status_code == 200:
                soup = BeautifulSoup(response.text, 'html.parser')
                for tag in soup(['script', 'style', 'nav', 'footer', 'head', 'link', 'meta', 'noscript']):
                    tag.decompose()

                main_content_areas = soup.find_all(['main', 'article', 'section', 'div'])
                if main_content_areas:
                    main_content = max(main_content_areas, key=lambda x: len(x.text))
                    content_tags = ['p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6']
                    content = ' '.join([tag.text.strip() for tag in main_content.find_all(content_tags)])
                else:
                    content = ' '.join([tag.text.strip() for tag in soup.find_all(['p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6'])])

                content = re.sub(r'\t', ' ', content)
                content = re.sub(r'\s+', ' ', content)
                return content
            elif response.status_code == 404:
                return f"Error: 404. Url is invalid or does not exist. Try with valid url..."
            else:
                logger.error(f"Error while extracting text from HTML (bs4): {response.status_code}")
                return f"Error while extracting text from HTML (bs4): {response.status_code}"

        except Exception as e:
            logger.error(f"Unknown error while extracting text from HTML (bs4): {str(e)}")
            return ""

    def extract_with_lxml(self, url):
        """
        Extract the text from a webpage using the lxml method.

        Args:
            url (str): The URL of the webpage to extract from.

        Returns:
            str: The extracted text.
        """
        try:
            config = Config()
            config.browser_user_agent = random.choice(USER_AGENTS)
            config.request_timeout = 10
            session = HTMLSession()

            response = session.get(url)
            response.html.render(timeout=config.request_timeout)
            html_content = response.html.html

            tree = html.fromstring(html_content)
            paragraphs = tree.cssselect('p, h1, h2, h3, h4, h5, h6')
            content = ' '.join([para.text_content() for para in paragraphs if para.text_content()])
            content = content.replace('\t', ' ').replace('\n', ' ').strip()

            return content

        except ArticleException as ae:
            logger.error("Error while extracting text from HTML (lxml): {str(ae)}")
            return ""

        except RequestException as re:
            logger.error(f"Error while making the request to the URL (lxml): {str(re)}")
            return ""

        except Exception as e:
            logger.error(f"Unknown error while extracting text from HTML (lxml): {str(e)}")
            return ""
    


# import asyncio
# import json
# import pytest
# from querent.callback.event_callback_interface import EventCallbackInterface
# from querent.common.types.querent_queue import QuerentQueue
# from querent.common.types.querent_event import EventState, EventType
# from querent.common.types.ingested_tokens import IngestedTokens
# from querent.common.types.querent_queue import QuerentQueue
# from querent.config.core.bert_llm_config import LLM_Config
# from querent.core.transformers.bert_llm import BERTLLM
# from querent.querent.resource_manager import ResourceManager
# from querent.querent.querent import Querent


# @pytest.mark.asyncio
# @pytest.mark.parametrize("input_data,ner_model_name, llm_class,expected_entities,filter_entities", [
#     #("Nishant is working from Delhi. Ansh is working from Punjab. Ayush is working from Odisha. India is very good at playing cricket. Nishant is working from Houston.", "dbmdz/bert-large-cased-finetuned-conll03-english", BERTLLM, ["http://geodata.org/Nishant", "http://geodata.org/Delhi"], False),
#     #("Nishant is working from Delhi. Ansh is working from Punjab. Ayush is working from Odisha. India is very good at playing cricket. Nishant is working from Houston.", "dbmdz/bert-large-cased-finetuned-conll03-english", BERTLLM, ["http://geodata.org/Nishant", "http://geodata.org/Delhi"], False),
#     ("""In this study, we present evidence of a Paleocene–Eocene Thermal Maximum (PETM)
# record within a 543-m-thick (1780 ft) deep-marine section in the Gulf of Mexico (GoM)
# using organic carbon stable isotopes and biostratigraphic constraints. We suggest that
# climate and tectonic perturbations in the upstream North American catchments can induce
# a substantial response in the downstream sectors of the Gulf Coastal Plain and ultimately
# in the GoM. This relationship is illustrated in the deep-water basin by (1) a high accom-
# modation and deposition of a shale interval when coarse-grained terrigenous material
# was trapped upstream at the onset of the PETM, and (2) a considerable increase in sedi-
# ment supply during the PETM, which is archived as a particularly thick sedimentary
# section in  the deep-sea fans of the GoM basin. The Paleocene–Eocene Thermal Maximum (PETM) (ca. 56 Ma) was a rapid global warming event characterized by the rise of temperatures to5–9 °C (Kennett and Stott, 1991), which caused substantial environmental changes around the globe.""",
# "botryan96/GeoBERT", BERTLLM, ["tectonic perturbations","downstream sectors"], True)])



# async def test_bertllm_ner_tokenization_and_entity_extraction(input_data, ner_model_name, llm_class, expected_entities, filter_entities):
#     input_queue = QuerentQueue()
#     resource_manager = ResourceManager()
#     ingested_data = IngestedTokens(file="dummy_1_file.txt", data=input_data)
#     await input_queue.put(ingested_data)
#     await input_queue.put(IngestedTokens(file="dummy_2_file.txt", data="dummy"))
#     await input_queue.put(IngestedTokens(file="dummy_2_file.txt", data=None, error="error"))
#     bert_llm_config = LLM_Config(
#         ner_model_name=ner_model_name,
#         enable_filtering=filter_entities,
#         filter_params={
#             'score_threshold': 0.5,
#             'attention_score_threshold': 0.1,
#             'similarity_threshold': 0.5,
#             'min_cluster_size': 5,
#             'min_samples': 3,
#             'cluster_persistence_threshold':0.2
#         }
#             ,fixed_entities = ['eocene', 'mexico']
#             , sample_entities=['B-GeoTime', 'B-GeoLoc']
#             , fixed_relationships=["constraint"]
#             , sample_relationships=["location", "locatedin"]
#             , is_confined_search= True
#     )
#     llm_instance = llm_class(input_queue, bert_llm_config)
#     class StateChangeCallback(EventCallbackInterface):
#         async def handle_event(self, event_type: EventType, event_state: EventState):
#             assert event_state.event_type == EventType.Graph
#             triple = json.loads(event_state.payload)
#             assert isinstance(triple['subject'], str) and triple['subject']
#     llm_instance.subscribe(EventType.Graph, StateChangeCallback())
#     querent = Querent(
#         [llm_instance],
#         resource_manager=resource_manager,
#     )

#     await querent.start()




from enum import Enum


class EngineConfigKey(Enum):
    ID = "id"
    NAME = "name"
    NUM_WORKERS = "num_workers"
    MAX_RETRIES = "max_retries"
    RETRY_INTERVAL = "retry_interval"
    MESSAGE_THROTTLE_LIMIT = "message_throttle_limit"
    MESSAGE_THROTTLE_DELAY = "message_throttle_delay"
    INNER_CHANNEL = "inner_channel"
    CHANNEL = "channel"



from querent.logging.logger import setup_logger

class FileBuffer:
    """
    A buffer to manage and store file content in chunks.

    This class handles file content that comes in chunks by storing these chunks in a dictionary.
    When a chunk with None is received, it is interpreted as the end of the current file,
    and the content of that file is processed.

    Attributes:
        current_filename (str): The name of the file currently being processed.
        file_chunks (dict): A dictionary storing file content chunks.

    Methods:
        add_chunk(filename: str, chunk: str) -> None:
            Add a chunk of data to the buffer for the current file. If chunk is None,
            process the end of the file.

        end_file(filename: str) -> str:
            Process the end of the specified file, clear its content from the dictionary,
            and return the full content of the file.

        get_content(filename: str) -> str:
            Retrieve the full content for a given filename from the dictionary.
    """

    def __init__(self):
        self.logger = setup_logger(__name__, "FileBuffer")
        self.file_chunks = {} 

    def add_chunk(self, filename, chunk):
        try:
            if chunk is None:
                return self.end_file(filename)

            if filename not in self.file_chunks:
                self.file_chunks[filename] = {}

            # Automatically assign a chunk ID
            chunk_id = len(self.file_chunks[filename])
            self.file_chunks[filename][chunk_id] = chunk

            return filename, None
                
        except Exception as e:
            self.logger.error(f"Error adding a chunk: {e}")
            raise Exception(f"An error occurred while adding a chunk: {e}")   

    def end_file(self, filename):
        try:
            # Assemble file content from chunks
            if filename in self.file_chunks:
                chunks = self.file_chunks[filename]
                full_content = ''.join([chunks[i] for i in sorted(chunks.keys())])
                del self.file_chunks[filename]  # Clear the file entry
                return filename, full_content
            else:
                raise Exception(f"No chunks found for file: {filename}")
        except Exception as e:
            self.logger.error(f"Error ending the file: {e}")
            raise Exception(f"An error occurred while ending the file: {e}")

    def get_content(self, filename):
        try:
            if filename in self.file_chunks:
                chunks = self.file_chunks[filename]
                return ''.join([chunks[i] for i in sorted(chunks.keys())])
            else:
                raise Exception(f"No content found for file: {filename}")
        except Exception as e:
            self.logger.error(f"Error getting content for {filename}: {e}")
            raise Exception(f"An error occurred while getting content for {filename}: {e}")



from typing import Union


class IngestedImages:
    def __init__(
        self,
        file: str,
        image: str,
        image_name: str,
        page_num: int,
        text: [str],
        coordinates: list = [],
        ocr_text: list = [],
        error: str = None,
    ) -> None:
        self.file = file
        self.text = text
        self.error = error
        self.image = image
        self.image_name = image_name
        self.page_num = page_num
        self.coordinates = coordinates
        self.ocr_text = ocr_text
        file = str(file)
        self.extension = file.split(".")[-1]
        self.file_id = file.split("/")[-1].split(".")[0]

    def __str__(self):
        if self.error:
            return f"Error: {self.error}"
        return f"Data: {self.ocr_text}"

    def is_error(self) -> bool:
        return self.error is not None

    def get_extension(self) -> str:
        return self.extension

    def get_file_id(self) -> str:
        return self.file_id



from querent.kg.ner_helperfunctions.ner_llm_transformer import NER_LLM
from querent.logging.logger import setup_logger
import torch
import torch.nn.functional as F


"""
EntityAttentionExtractor Class:

This class is designed to extract attention weights from a given transformer model for specific entities within a context. 
The primary purpose is to determine the significance or relevance of entities in the context based on the attention mechanism of the transformer model.

Attributes:
    model (torch.nn.Module): The transformer model used for extracting attention weights.
    tokenizer (Tokenizer): The tokenizer associated with the transformer model.

Methods:
    extract_attention_weight(entity, context):
        Extracts the attention weight for a given entity within a context.
        
    extract_and_append_attention_weights(doc_entity_pairs):
        Processes a list of document-entity pairs, extracts attention weights for each entity, 
        and appends the attention weights to the respective pairs.

Dependencies:
    - torch
    - NER_LLM from querent.kg.ner_helperfunctions.ner_llm_transformer
    - setup_logger from querent.logging.logger

Example:
    model = ... # Load a transformer model
    tokenizer = ... # Load the corresponding tokenizer
    extractor = EntityAttentionExtractor(model, tokenizer)
    attention_weight = extractor.extract_attention_weight("apple", "Apple Inc. is a tech company.")
    ...

Note:
    Ensure that the transformer model and tokenizer are compatible. 
    The model should be capable of returning attention weights (e.g., BERT, RoBERTa).

"""

class EntityAttentionExtractor:
    def __init__(self, model, tokenizer):
        self.model = model
        self.tokenizer = tokenizer
        self.logger = setup_logger(__name__, "EntityAttentionExtractor")

    def extract_attention_weight(self, entity, context):
        try:
            inputs = self.tokenizer(context, return_tensors="pt", truncation=True, padding=True, max_length=512)
            with torch.no_grad():
                outputs = self.model(**inputs, output_attentions=True)
                attentions = outputs.attentions[-1][0] 
            entity_token_ids = self.tokenizer.encode(entity, add_special_tokens=False)
            entity_positions = [i for i, token_id in enumerate(inputs["input_ids"][0]) if token_id in entity_token_ids]
            
            non_zero_attentions = []
            # Iterate over each head's attentions
            for head in range(attentions.size(0)):
                # Extract the attention weights for the entity tokens
                head_attentions = attentions[head, :, :][:, entity_positions]
                # Flatten the attention weights and filter out zeros
                head_attentions = head_attentions.flatten().tolist()
                non_zero_attentions.extend([att for att in head_attentions if att > 0])

            # Convert to a tensor for further computation
            non_zero_attentions = torch.tensor(non_zero_attentions)

            # Calculate the weighted mean of non-zero attention weights
            if len(non_zero_attentions) > 0:
                weighted_mean_attention = (non_zero_attentions * non_zero_attentions).sum() / non_zero_attentions.sum()
                attention_weight = weighted_mean_attention.item()
            else:
                attention_weight = 0
            
            return attention_weight
        
        except Exception as e:
            self.logger.error(f"Error extracting attention weights : {e}")
            raise Exception("Error extracting attention weights : {}".format(e))

    def extract_and_append_attention_weights(self, doc_entity_pairs):
        updated_pairs = []
        try:
            for inner_list in doc_entity_pairs:
                updated_inner_list = []
                for pair in inner_list:
                    full_context = pair[1]
                    entity1 = pair[0]
                    entity2 = pair[2]
                    sentences = NER_LLM.split_into_sentences(full_context)
                    context = None
                    for sentence in sentences:
                        if entity1.lower() in sentence.lower() and entity2.lower() in sentence.lower():
                            context = sentence
                            break
                    if not context:
                        context = full_context
                    entity1_attnscore = self.extract_attention_weight(entity1, context)
                    entity2_attnscore = self.extract_attention_weight(entity2, context)
                    pair_dict = pair[3]
                    pair_dict['entity1_attnscore'] = round(entity1_attnscore,2)
                    pair_dict['entity2_attnscore'] = round(entity2_attnscore,2)
                    # harmonic mean for a pair- useful to penalize entity pairs where one entity has a much lower score than the other.
                    if entity1_attnscore > 0 and entity2_attnscore > 0:
                        pair_dict['pair_attnscore'] = round(2.0 * (entity1_attnscore * entity2_attnscore) / 
                                                            (entity1_attnscore + entity2_attnscore), 2)
                    else:
                        pair_dict['pair_attnscore'] = 0

                    updated_inner_list.append((entity1, full_context, entity2, pair_dict))
                updated_pairs.append(updated_inner_list)
                
            return updated_pairs
        
        except Exception as e:
            self.logger.error(f"Error extracting and appending attention weights : {e}")
            raise Exception("Error extracting and appending attention weights : {}".format(e))



from pydantic import BaseModel, Field
from typing import Dict

class Opensource_LLM_Config(BaseModel):
    name: str = "RelationshipExtractor"
    description: str = "An engine for extracting relationships"
    version: str = "0.0.1"
    logger: str = "RelationshipExtractor.engine_config"
    model_type: str = 'llama'
    model_path: str = './tests/llama-2-7b-chat.Q5_K_M.gguf' 
    grammar_file_path: str = './querent/kg/rel_helperfunctions/json.gbnf'
    qa_template: str = Field(default=None)
    vector_store_path: str = "./querent/kg/rel_helperfunctions/vectorstores/"
    emb_model_name: str = 'sentence-transformers/all-MiniLM-L6-v2'
    rag_approach: bool = False

    def __init__(self, config_source=None, **kwargs):
        declared_keys = set(self.__annotations__.keys())

        # Initialize BaseModel with kwargs
        super().__init__(**kwargs)

        if config_source:
            config_data = self.load_config(config_source)
            for config_key, config_value in config_data.items():
                if config_key in declared_keys:
                    setattr(self, config_key, config_value)

        # Apply any additional keyword arguments, if they are declared attributes
        for key, value in kwargs.items():
            if key in declared_keys:
                setattr(self, key, value)

    @staticmethod
    def load_config(config_source):
        if isinstance(config_source, dict):
            return config_source
        else:
            raise ValueError("Invalid config. Must be a valid dictionary")

    def get_faiss_index_path(self):
        return self.vector_store_path + "my_FAISS_index"


import json
import pytest
from querent.config.graph_config import GraphConfig
from querent.graph.subject import Subject
from querent.kg.querent_kg import QuerentKG
from querent.graph.utils import URI, BNode, Literal


@pytest.fixture
def schema():
    # Set up a sample ontology schema
    return json.dumps(
        {
            "fields": {
                "name": {"type": "string", "description": "Name of the person"},
                "location": {"type": "string", "description": "Location of the person"},
            }
        }
    )


@pytest.fixture
def querent_kg(schema):
    # Set up a QuerentKG instance for testing
    config = GraphConfig(
        identifier="test_kg",
        format="nt",
        store="default",
        schema=schema,
        flush_on_serialize=False,
    )
    return QuerentKG(config)


def test_add_knowledge(querent_kg: QuerentKG):
    # Create a Subject instance to represent knowledge about Alice
    alice = Subject(URI("http://example.org/Alice"))

    # Define properties for Alice
    alice.add_property(URI("http://example.org/hasLocation"), Literal("Paris"))

    # Add Alice's data to the knowledge graph
    querent_kg.add_subject_knowledge([alice])

    # Ensure that the knowledge graph is not empty
    serialized_data = querent_kg.serialize()
    assert serialized_data != ""

    # Verify that the correct data has been added
    triples = querent_kg.graph.triples((None, None, None))
    assert len(list(triples)) == 1

    # check the memory of graph
    expect_bytes = 119
    assert querent_kg.get_current_memory_usage == expect_bytes



import asyncio
from asyncio import Queue
from pathlib import Path
from querent.collectors.fs.fs_collector import FSCollectorFactory
from querent.config.collector.collector_config import FSCollectorConfig
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
import pytest
import uuid


@pytest.mark.asyncio
async def test_ingest_all_async():
    # Set up the collectors
    directories = ["./tests/data/pdf/", "./tests/data/pdf2/", "./tests/data/pdf3/"]
    collectors = [
        FSCollectorFactory().resolve(
            Uri("file://" + str(Path(directory).resolve())),
            FSCollectorConfig(
                config_source={
                    "id": str(uuid.uuid4()),
                    "root_path": directory,
                    "name": "Local-config",
                    "config": {},
                    "uri": "file://",
                }
            ),
        )
        for directory in directories
    ]

    # Set up the result queue
    result_queue = asyncio.Queue()

    # Create the IngestorFactoryManager
    ingestor_factory_manager = IngestorFactoryManager(
        collectors=collectors, result_queue=result_queue
    )

    # Start the ingest_all_async in a separate task
    ingest_task = asyncio.create_task(ingestor_factory_manager.ingest_all_async())

    # Wait for the task to complete
    await asyncio.gather(ingest_task)

    # Optionally, check the result_queue for ingested data
    counter = 0
    unique_files = set()
    while not result_queue.empty():
        ingested_data = await result_queue.get()
        if ingested_data is not None:
            unique_files.add(ingested_data.file)
            counter += 1
    print(f"Found {counter} ingested files")
    assert counter == 98
    assert len(unique_files) == 5


if __name__ == "__main__":
    asyncio.run(test_ingest_all_async())



import asyncio
import json
from querent.collectors.collector_resolver import CollectorResolver
from querent.collectors.gcs.gcs_collector import GCSCollectorFactory
from querent.common.uri import Uri
from querent.config.collector.collector_config import CollectorBackend, GcsCollectConfig
import pytest
import uuid
from dotenv import load_dotenv

load_dotenv()


@pytest.fixture
def gcs_config():
    cred_file = "/tmp/.config/gcloud/application_default_credentials.json"
    credentials_info = json.load(open(cred_file))
    credential_json_str = json.dumps(credentials_info)
    return GcsCollectConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "bucket": "querent-test",
            "credentials": credential_json_str,
            "chunk": 1024,
            "config": {},
            "name": "GCS-config",
            "uri": "gs://",
        }
    )


def test_gcs_collector_factory():
    factory = GCSCollectorFactory()
    assert factory.backend() == CollectorBackend.Gcs


# Modify this function to test the GCS collector

# To do: uncomment the following code when you have the bucket name and the credentials.json file for testing.


@pytest.mark.asyncio
async def test_gcs_collector(gcs_config):
    config = gcs_config
    uri = Uri("gs://" + config.bucket)
    resolver = CollectorResolver()
    collector = resolver.resolve(uri, config)
    assert collector is not None

    await collector.connect()

    async def poll_and_print():
        counter = 0
        async for result in collector.poll():
            assert not result.is_error()
            chunk = result.unwrap()

            if chunk is not None:
                counter += 1
        assert counter == 797

    await poll_and_print()


if __name__ == "__main__":
    asyncio.run(test_gcs_collector())



import asyncio

from querent.config.collector.collector_config import S3CollectConfig
from querent.collectors.collector_resolver import CollectorResolver
from querent.collectors.aws.aws_collector import AWSCollectorFactory
from querent.common.uri import Uri
from querent.config.collector.collector_config import CollectorBackend
import pytest
import os
from dotenv import load_dotenv
import uuid

load_dotenv()


aws_access_key_id = os.getenv("AWS_ACCESS_KEY_ID")
aws_secret_access_key = os.getenv("AWS_SECRET_ACCESS_KEY")


@pytest.fixture
def aws_config():
    config = S3CollectConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "bucket": "pstreamsbucket1",
            "region": "ap-south-1",
            "access_key": aws_access_key_id,
            "secret_key": aws_secret_access_key,
            "chunk": 1024,
            "name": "AWS-config",
            "config": {},
            "backend": "s3",  # Assuming this is a string or relevant enum value
            "uri": "s3://",
        },
    )
    return config


def test_aws_collector_factory():
    factory = AWSCollectorFactory()
    assert factory.backend() == CollectorBackend.S3


# Modify this function to test the AWS collector


@pytest.mark.asyncio
async def test_aws_collector(aws_config):
    config = aws_config
    uri = Uri("s3://" + config.bucket)
    resolver = CollectorResolver()
    collector = resolver.resolve(uri, config)
    assert collector is not None

    await collector.connect()

    async def poll_and_print():
        counter = 0
        async for result in collector.poll():
            assert not result.is_error()
            chunk = result.unwrap()

            if chunk is not None:
                counter += 1
        assert counter == 5392

    await poll_and_print()


if __name__ == "__main__":
    asyncio.run(test_aws_collector())



from typing import List, AsyncGenerator

from querent.ingestors.base_ingestor import BaseIngestor
from querent.ingestors.ingestor_factory import IngestorFactory
from querent.processors.async_processor import AsyncProcessor
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.common.types.ingested_code import IngestedCode
from querent.common.types.collected_bytes import CollectedBytes


class GithubIngestor(BaseIngestor):
    """Class for ingesting github code"""

    def __init__(self, processors: List[AsyncProcessor]):
        super().__init__(IngestorBackend.GITHUB)
        self.processorrs = processors

    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedCode, None]:
        collected_bytes = b""
        current_file = None
        try:
            async for chunk_bytes in poll_function:
                if chunk_bytes.is_error() or chunk_bytes.is_eof():
                    continue

                if current_file is None:
                    current_file = chunk_bytes.file
                elif current_file != chunk_bytes.file:
                    async for line in self.extract_and_process_code(
                        CollectedBytes(file=current_file, data=collected_bytes)
                    ):
                        yield IngestedCode(
                            file=current_file,
                            data=[line],  # Wrap line in a list
                            error=None,
                        )
                    yield IngestedCode(
                        file=current_file,
                        data=None,
                        error=None,
                    )
                    collected_bytes = b""
                    current_file = chunk_bytes.file

                collected_bytes += chunk_bytes.data

            if current_file:
                async for line in self.extract_and_process_code(
                    CollectedBytes(file=current_file, data=collected_bytes)
                ):
                    yield IngestedCode(
                        file=current_file,
                        data=[line],  # Wrap line in a list
                        error=None,
                    )
                yield IngestedCode(file=current_file, data=None, error=None)
        except Exception as e:
            yield IngestedCode(file=current_file, data=None, error=f"Exception: {e}")

    async def extract_and_process_code(
        self, chunk_bytes: CollectedBytes
    ) -> AsyncGenerator[IngestedCode, None]:
        message_bytes = chunk_bytes.data.decode("UTF-8")
        yield message_bytes


class GithubIngestorFactory(IngestorFactory):
    """Class for github ingestor factory"""

    SUPPORTED_EXTENSIONS = {"github"}

    async def supports(self, file_extension: str) -> bool:
        return file_extension.lower() in self.SUPPORTED_EXTENSIONS

    async def create(
        self, file_extension: str, processors: List[AsyncProcessor]
    ) -> BaseIngestor:
        if not await self.supports(file_extension):
            return None
        return GithubIngestor(processors)



import asyncio
from pathlib import Path
import tempfile
from querent.collectors.collector_resolver import CollectorResolver
from querent.collectors.webscaper.web_scraper_collector import WebScraperFactory
import pytest
import uuid

from querent.common.uri import Uri
from querent.config.collector.collector_config import CollectorBackend, WebScraperConfig


def test_webscrapper_collector():
    uri = Uri("https://asecuritysite.com/")
    resolver = CollectorResolver()
    webscrapperConfig = WebScraperConfig(
        config_source={
            "website_url": uri.uri,
            "id": str(uuid.uuid4()),
            "name": "Webscrapper-config",
            "config": {},
            "uri": "https://asecuritysite.com/",
        }
    )
    collector = resolver.resolve(uri, webscrapperConfig)
    assert collector is not None


def test_fs_collector_factory():
    factory = WebScraperFactory()
    assert factory.backend() == CollectorBackend.WebScraper


def test_scrapping_data():
    uri = Uri("https://protocolstreams.xyz/")
    resolver = CollectorResolver()
    webscrapperConfig = WebScraperConfig(
        config_source={
            "website_url": uri.uri,
            "id": str(uuid.uuid4()),
            "name": "Webscrapper-config",
            "config": {},
            "uri": "https://protocolstreams.xyz/",
        }
    )
    collector = resolver.resolve(uri, webscrapperConfig)
    assert collector is not None

    async def poll_and_print():
        async for result in collector.poll():
            assert not result.is_error()

    asyncio.run(poll_and_print())



import asyncio
import imaplib
from typing import AsyncGenerator

from querent.collectors.collector_base import Collector
from querent.collectors.collector_factory import CollectorFactory
from querent.collectors.email.imap import ImapEmail
from querent.common import common_errors
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.uri import Uri
from querent.config.collector.collector_config import (
    CollectorBackend,
    EmailCollectorConfig,
)
from querent.logging.logger import setup_logger


class EmailCollector(Collector):
    def __init__(self, config: EmailCollectorConfig):
        self.config = config
        self.imap_email = ImapEmail()
        self.logger = setup_logger(__name__, "EmailCollector")

    async def connect(self):
        try:
            # Open an IMAP connection using the provided configuration
            self.imap_connection = self.imap_email.imap_open(self.config)
            # Other connection-related setup can be done here if needed

        except imaplib.IMAP4.error as e:
            self.logger.error(f"Error connecting to IMAP server: {e}")
            raise common_errors.ConnectionError(
                "Failed to connect to the IMAP server: {}".format(str(e))
            ) from e

    async def disconnect(self):
        try:
            if self.imap_connection:
                # Close the IMAP connection
                await asyncio.get_event_loop().run_in_executor(
                    None, self.imap_connection.logout
                )
        except Exception as e:
            self.logger.error(f"Error disconnecting from IMAP server: {e}")

    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        try:
            _, messages = self.imap_connection.select(self.config.imap_folder)
            num_of_messages = int(messages[0])

            # Fetch the emails
            for i in range(num_of_messages, 0, -1):
                _, data = self.imap_connection.fetch(str(i), "(RFC822)")
                for response_part in data:
                    if isinstance(response_part, tuple):
                        message = response_part[1]
                        yield CollectedBytes(
                            data=message,
                            file=f"{self.config.imap_username}:{self.config.imap_folder}/{i}.email",
                        )
                yield CollectedBytes(
                    data=None,
                    file=f"{self.config.imap_username}:{self.config.imap_folder}/{i}.email",
                    eof=True,
                )
        except imaplib.IMAP4.error as e:
            self.logger.error(f"Error fetching emails from IMAP server: {e}")
            raise common_errors.ConnectionError(
                "Failed to fetch emails from the IMAP server: {}".format(str(e))
            ) from e
        finally:
            await self.disconnect()


class EmailCollectorFactory(CollectorFactory):
    def backend(self) -> CollectorBackend:
        return CollectorBackend.Email

    def resolve(self, uri: Uri, config: EmailCollectorConfig) -> Collector:
        return EmailCollector(config)



import json
import numpy as np
import hdbscan
from scipy.spatial.distance import cosine
from sklearn.preprocessing import StandardScaler
from sklearn.metrics.pairwise import cosine_similarity
from querent.logging.logger import setup_logger
from typing import List, Tuple, Dict

"""
    A class used to filter and cluster triples of (entity1, context, entity2) based on various scoring metrics and embedding similarities.

    Attributes
    ----------
    score_threshold : float
        The minimum score that both entities in a triple must have to be considered relevant.
    attention_score_threshold : float
        The minimum attention score that a triple must have to be considered relevant.
    similarity_threshold : float
        The minimum cosine similarity score between the embeddings of the entities to be considered relevant.
    min_cluster_size : int
        The minimum size of clusters to be formed by the HDBSCAN clustering algorithm.
    min_samples : int
        The number of samples in a neighborhood for a point to be considered a core point.

    Methods
    -------
    calculate_cosine_similarity(embedding1, embedding2):
        Calculates the cosine similarity between two embeddings.

    filter_by_score(data):
        Filters triples by checking if the entity scores meet the score threshold.

    filter_by_attention_score(data):
        Filters triples by checking if the pair attention score meets the attention score threshold.

    filter_by_embedding_similarity(data):
        Filters triples by checking if the cosine similarity between entity embeddings meets the similarity threshold.

    filter_triples(triples):
        Filters a list of triples by applying score, attention score, and embedding similarity filters.

    combine_embeddings(entity1_embedding, entity2_embedding):
        Combines two entity embeddings into a single embedding.

    cluster_triples(triples):
        Clusters the filtered triples using the HDBSCAN algorithm and returns the clusters along with the reduction count and the clusterer instance.
    
    filter_by_cluster_persistence:
        Filter triples by cluster persistence to retain only those belonging to clusters with high persistence values.
        
    Notes
    -----
    - The class expects triples in the form of (entity1, context, entity2), where 'context' is a JSON string containing various attributes including the embeddings.
    - The embeddings are expected to be in a format that can be converted to numpy arrays.
    - The class is designed to be flexible and allows for customization of the clustering process through various parameters.
"""

class TripleFilter:
    def __init__(self, score_threshold, attention_score_threshold, similarity_threshold, min_cluster_size=2, min_samples=None, cluster_persistence_threshold=-1):
        self.score_threshold = score_threshold
        self.attention_score_threshold = attention_score_threshold
        self.similarity_threshold = similarity_threshold
        self.min_cluster_size = min_cluster_size
        self.logger = setup_logger(__name__, "TripleFilter")
        self.min_samples = min_samples if min_samples is not None else self.min_cluster_size
        self.cluster_persistence_threshold = cluster_persistence_threshold

    @staticmethod
    def calculate_cosine_similarity(embedding1: np.ndarray, embedding2: np.ndarray) -> float:
        embedding1 = np.array(embedding1)
        embedding2 = np.array(embedding2)
        
        return cosine_similarity(embedding1.reshape(1, -1), embedding2.reshape(1, -1))[0, 0]

    def filter_by_score(self, data: dict) -> bool:
        
        return data['entity1_score'] >= self.score_threshold and data['entity2_score'] >= self.score_threshold

    def filter_by_attention_score(self, data: dict) -> bool:
        
        return data['pair_attnscore'] >= self.attention_score_threshold

    def filter_by_embedding_similarity(self, data: dict) -> bool:
        similarity = self.calculate_cosine_similarity(data['entity1_embedding'], data['entity2_embedding'])
        
        return similarity >= self.similarity_threshold

    def filter_triples(self, triples: List[Tuple[str, str, str]]) -> Tuple[List[Tuple[str, str, str]], int]:
        relevant_triples = []
        initial_count = len(triples)
        for triple in triples:
            try:
                entity1, json_data, entity2 = triple
                data = json.loads(json_data)                
                if not self.filter_by_score(data) or not self.filter_by_attention_score(data) or not self.filter_by_embedding_similarity(data):
                    continue

                relevant_triples.append(triple)
            except (json.JSONDecodeError, KeyError) as e:
                self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to filter triples {e}")
                raise Exception(f"Unable to filter triples: {e}")
        reduction_count = initial_count - len(relevant_triples)
        
        return relevant_triples, reduction_count

    @staticmethod
    def combine_embeddings(entity1_embedding: np.ndarray, entity2_embedding: np.ndarray) -> np.ndarray:
        return np.concatenate((entity1_embedding, entity2_embedding))

    def cluster_triples(self, triples: List[Tuple[str, str, str]]) -> Dict[str, any]:
        try:
            combined_embeddings = np.array([
                self.combine_embeddings(
                    json.loads(triple[1])['entity1_embedding'],
                    json.loads(triple[1])['entity2_embedding']
                ) for triple in triples
            ])
            scaler = StandardScaler()
            normalized_embeddings = scaler.fit_transform(combined_embeddings)
            distance_matrix = 1 - cosine_similarity(normalized_embeddings)
            clusterer = hdbscan.HDBSCAN(
                metric='precomputed',
                min_cluster_size=self.min_cluster_size,
                min_samples=self.min_samples,
                cluster_selection_method="leaf"
            )
            cluster_labels = clusterer.fit_predict(distance_matrix)
            cluster_persistence = clusterer.cluster_persistence_

            filtered_triples = [triples[index] for index, label in enumerate(cluster_labels) if label != -1]
            cluster_output = {
                'filtered_triples': filtered_triples,
                'reduction_count': len(triples) - len(filtered_triples),
                'cluster_labels': cluster_labels,
                'cluster_persistence': cluster_persistence
            }

            return cluster_output

        except Exception as e:
            self.logger.error(f"Invalid {self.__class__.__name__} configuration. Unable to apply clustering on triples {e}")
            raise Exception(f"Error during clustering: {e}")

    def filter_by_cluster_persistence(self, triples: List[Tuple[str, str,str]], cluster_persistence, cluster_labels) -> List[Tuple[str, str, str]]:
        if self.cluster_persistence_threshold != -1:
            high_persistence_triples = []
            high_persistence_clusters = [index for index, persistence in enumerate(cluster_persistence) if persistence > self.cluster_persistence_threshold]
            for i, (triple, label) in enumerate(zip(triples, cluster_labels)):
                if label in high_persistence_clusters:
                    high_persistence_triples.append(triple)
                    
            return high_persistence_triples
        
        else:
            
            return None


import asyncio
from pathlib import Path
import tempfile
from querent.collectors.collector_resolver import CollectorResolver
from querent.collectors.fs.fs_collector import FSCollectorFactory
import pytest
import uuid

from querent.common.uri import Uri
from querent.config.collector.collector_config import (
    CollectorBackend,
    FSCollectorConfig,
)


@pytest.fixture
def temp_dir():
    temp_dir = tempfile.TemporaryDirectory()
    yield temp_dir.name
    temp_dir.cleanup()


def test_fs_collector(temp_dir):
    uri = Uri("file://" + temp_dir)
    resolver = CollectorResolver()
    fileConfig = FSCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "root_path": uri.path,
            "name": "Local-config",
            "config": {},
            "uri": "file://",
        }
    )
    collector = resolver.resolve(uri, fileConfig)
    assert collector is not None


def test_fs_collector_factory():
    factory = FSCollectorFactory()
    assert factory.backend() == CollectorBackend.LocalFile


async def poll_and_print(collector):
    async for result in collector.poll():
        assert not result.is_error()
        chunk = result.unwrap()


async def add_files(temp_dir):
    file_path = Path(temp_dir, "test_temp.txt")
    with open(file_path, "wb") as file:
        file.write(b"test_add_files_read_via_collector")


async def main():
    temp_dir = tempfile.TemporaryDirectory()
    uri = Uri("file://" + temp_dir.name)
    resolver = CollectorResolver()
    fileConfig = FSCollectorConfig(
        config_source={
            "id": str(uuid.uuid4()),
            "root_path": uri.path,
            "name": "Local-config",
            "config": {},
            "uri": "file://",
        }
    )
    collector = resolver.resolve(uri, fileConfig)
    assert collector is not None

    await add_files(temp_dir.name)
    await poll_and_print(collector)

    temp_dir.cleanup()


if __name__ == "__main__":
    asyncio.run(main())



from enum import Enum
from typing import Optional


class IngestorBackend(str, Enum):
    PDF = "pdf"
    TEXT = "txt"
    DOCX = "docx"
    DOC = "doc"
    PPT = "ppt"
    PPTX = "pptx"
    CSV = "csv"
    XLSX = "xlsx"
    JSON = "json"
    XML = "xml"
    HTML = "html"
    YAML = "yaml"
    MARKDOWN = "markdown"
    IMG = "image"
    PNG = "png"
    JPG = "jpg"
    GIF = "gif"
    WEBRTC = "webrtc"
    MP3 = "mp3"
    MP4 = "mp4"
    MOV = "mov"
    AVI = "avi"
    WAV = "wav"
    GITHUB = "github"
    Unsupported = "unsupported"
    Email = "email"
    Slack = "slack"
    Jira = "jira"
